# -*- coding: utf-8 -*-
"""
Automatic tests for class ldap0.schema.SubSchema
"""

import unittest

import ldap0.ldif
import ldap0.schema.util
from ldap0.schema.models import SchemaElement
from ldap0.schema.models import AttributeType
from ldap0.schema.models import ObjectClass
from ldap0.schema.models import LDAPSyntax
from ldap0.schema.models import MatchingRule
from ldap0.schema.models import MatchingRuleUse
from ldap0.schema.models import DITContentRule
from ldap0.schema.models import DITStructureRule
from ldap0.schema.models import NameForm
from ldap0.schema.models import Entry


# class attribute names of ObjectClass instances
OC_CLASS_ATTRS = ()


class SchemaDescriptionTest(unittest.TestCase):
    """
    Base class for testing parsing and unparsing of schema description classes
    """
    schema_class = SchemaElement
    class_attrs = (
        'oid',
        'desc',
    )
    test_cases = (
        (
            "( 1.2.3.4.5.6.7.8 DESC 'something weird: äöüÄÖÜß' )",
            {
                'oid': '1.2.3.4.5.6.7.8',
                'desc': 'something weird: äöüÄÖÜß',
            },
        ),
    )

    def test_parse(self):
        for schema_desc_str, res in self.test_cases:
            if not schema_desc_str:
                # skip empty strings
                continue
            at = self.schema_class(schema_desc_str)
            for class_attr in self.class_attrs:
                self.assertEqual(
                    getattr(at, class_attr),
                    res.get(class_attr, None),
                    'Test-case %r class attribute %r: %r != %r' % (
                        schema_desc_str,
                        class_attr,
                        getattr(at, class_attr),
                        res.get(class_attr, None),
                    ),
                )

    def test_unparse(self):
        for schema_desc_str, _ in self.test_cases:
            if not schema_desc_str:
                # skip empty strings
                continue
            at1 = self.schema_class(schema_desc_str)
            at2 = self.schema_class(str(at1))
            for class_attr in self.class_attrs:
                self.assertEqual(
                    getattr(at1, class_attr),
                    getattr(at2, class_attr),
                    'Test-case %r class attribute %r: %r != %r' % (
                        schema_desc_str,
                        class_attr,
                        getattr(at1, class_attr),
                        getattr(at2, class_attr),
                    ),
                )


class TestAttributeType(SchemaDescriptionTest):
    """
    test ldap0.schema.models.AttributeType
    """
    schema_class = AttributeType
    class_attrs = (
        'oid',
        'names',
        'desc',
        'obsolete',
        'sup',
        'equality',
        'ordering',
        'substr',
        'x_origin',
        'x_ordered',
        'syntax',
        'syntax_len',
        'single_value',
        'collective',
        'no_user_mod',
        'usage',
    )
    test_cases = (
        (
            "( 2.5.4.3 NAME ( 'cn' 'commonName' ) DESC 'RFC4519: common name(s) for which the entity is known by' SUP name )",
            {
                'oid': '2.5.4.3',
                'names': ('cn', 'commonName'),
                'desc': 'RFC4519: common name(s) for which the entity is known by',
                'obsolete': False,
                'sup': ('name',),
                'single_value': False,
                'collective': False,
                'no_user_mod': False,
                'usage': 0,
            },
        ),
        (
            "( 1.3.6.1.1.1.1.0 NAME 'uidNumber' DESC 'RFC2307: An integer uniquely identifying a user in an administrative domain' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
            {
                'oid': '1.3.6.1.1.1.1.0',
                'names': ('uidNumber',),
                'desc': 'RFC2307: An integer uniquely identifying a user in an administrative domain',
                'obsolete': False,
                'sup': (),
                'equality': 'integerMatch',
                'ordering': 'integerOrderingMatch',
                'syntax': '1.3.6.1.4.1.1466.115.121.1.27',
                'single_value': True,
                'collective': False,
                'no_user_mod': False,
                'usage': 0,
            },
        ),
        (
            "( 1.3.6.1.4.1.4203.1.12.2.3.2.3.2 NAME 'olcDbACLAuthcDn' DESC 'Remote ACL administrative identity' OBSOLETE SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 SINGLE-VALUE )",
            {
                'oid': '1.3.6.1.4.1.4203.1.12.2.3.2.3.2',
                'names': ('olcDbACLAuthcDn',),
                'desc': 'Remote ACL administrative identity',
                'obsolete': True,
                'sup': (),
                'syntax': '1.3.6.1.4.1.1466.115.121.1.12',
                'single_value': True,
                'collective': False,
                'no_user_mod': False,
                'usage': 0,
            },
        ),
        (
            "( 2.5.18.1 NAME 'createTimestamp' DESC 'RFC4512: time which object was created' EQUALITY generalizedTimeMatch ORDERING generalizedTimeOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.24 SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )",
            {
                'oid': '2.5.18.1',
                'names': ('createTimestamp',),
                'desc': 'RFC4512: time which object was created',
                'obsolete': False,
                'sup': (),
                'equality': 'generalizedTimeMatch',
                'ordering': 'generalizedTimeOrderingMatch',
                'syntax': '1.3.6.1.4.1.1466.115.121.1.24',
                'single_value': True,
                'collective': False,
                'no_user_mod': True,
                'usage': 1,
            },
        ),
        (
            "( 1.3.6.1.4.1.4203.1.12.2.3.0.33 NAME 'olcObjectIdentifier' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 X-ORDERED 'VALUES' )",
            {
                'oid': '1.3.6.1.4.1.4203.1.12.2.3.0.33',
                'names': ('olcObjectIdentifier',),
                'desc': None,
                'obsolete': False,
                'sup': (),
                'equality': 'caseIgnoreMatch',
                'substr': 'caseIgnoreSubstringsMatch',
                'ordering': None,
                'syntax': '1.3.6.1.4.1.1466.115.121.1.15',
                'single_value': False,
                'collective': False,
                'no_user_mod': False,
                'no_user_mod': False,
                'x_ordered': 'VALUES',
                'usage': 0,
            },
        ),
    )


class TestObjectClass(SchemaDescriptionTest):
    """
    test ldap0.schema.models.ObjectClass
    """
    schema_class = ObjectClass
    class_attrs = (
        'oid',
        'names',
        'obsolete',
        'desc',
        'must',
        'may',
        'kind',
        'sup',
    )
    test_cases = (
        (
            "( 2.16.840.1.113730.3.2.2 NAME 'inetOrgPerson' DESC 'RFC2798: Internet Organizational Person' SUP organizationalPerson STRUCTURAL MAY ( audio $ businessCategory $ carLicense $ departmentNumber $ displayName $ employeeNumber $ employeeType $ givenName $ homePhone $ homePostalAddress $ initials $ jpegPhoto $ labeledURI $ mail $ manager $ mobile $ o $ pager $ photo $ roomNumber $ secretary $ uid $ userCertificate $ x500uniqueIdentifier$ preferredLanguage $ userSMIMECertificate $ userPKCS12 ) )",
            {
                'oid': '2.16.840.1.113730.3.2.2',
                'names': ('inetOrgPerson', ),
                'obsolete': False,
                'desc': 'RFC2798: Internet Organizational Person',
                'must': (),
                'may': (
                    'audio', 'businessCategory', 'carLicense',
                    'departmentNumber', 'displayName', 'employeeNumber',
                    'employeeType', 'givenName', 'homePhone',
                    'homePostalAddress', 'initials', 'jpegPhoto',
                    'labeledURI', 'mail', 'manager', 'mobile', 'o',
                    'pager', 'photo', 'roomNumber', 'secretary', 'uid',
                    'userCertificate', 'x500uniqueIdentifier',
                    'preferredLanguage', 'userSMIMECertificate',
                    'userPKCS12',
                ),
                'kind': 0,
                'sup': ('organizationalPerson',),
            },
        ),
        (
            "( 1.2.3.4 NAME ( foo $ bar ) DESC 'äöüÄÖÜß' ABSTRACT OBSOLETE )",
            {
                'oid': '1.2.3.4',
                'names': ('foo', 'bar'),
                'obsolete': True,
                'desc': 'äöüÄÖÜß',
                'must': (),
                'may': (),
                'kind': 1,
                'sup': (),
            },
        ),
        (
            "( 1.2.3.4 NAME ( foobar ) AUXILIARY OBSOLETE )",
            {
                'oid': '1.2.3.4',
                'names': ('foobar', ),
                'obsolete': True,
                'must': (),
                'may': (),
                'kind': 2,
                'sup': (),
            },
        ),
    )


class TestLDAPSyntax(SchemaDescriptionTest):
    """
    test ldap0.schema.models.LDAPSyntax
    """
    schema_class = LDAPSyntax
    class_attrs = (
        'oid',
        'desc',
        'x_subst',
        'not_human_readable',
        'x_binary_transfer_required',
    )
    test_cases = (
        (
            "( 1.3.6.1.4.1.1466.115.121.1.4 DESC 'Audio' X-NOT-HUMAN-READABLE 'TRUE' )",
            {
                'oid': '1.3.6.1.4.1.1466.115.121.1.4',
                'desc': 'Audio',
                'not_human_readable': True,
                'x_binary_transfer_required': False,
            },
        ),
        (
            "( 1.3.6.1.4.1.1466.115.121.1.6 DESC 'Bit String' )",
            {
                'oid': '1.3.6.1.4.1.1466.115.121.1.6',
                'desc': 'Bit String',
                'not_human_readable': False,
                'x_binary_transfer_required': False,
            },
        ),
        (
            "( 1.3.6.1.4.1.1466.115.121.1.8 DESC 'Certificate' X-BINARY-TRANSFER-REQUIRED 'TRUE' X-NOT-HUMAN-READABLE 'TRUE' )",
            {
                'oid': '1.3.6.1.4.1.1466.115.121.1.8',
                'desc': 'Certificate',
                'x_subst': None,
                'not_human_readable': True,
                'x_binary_transfer_required': True,
            },
        ),
        (
            "( 1.2.3.4 DESC 'Weirdest test äöüÄÖÜß' X-SUBST blurbBlabber X-BINARY-TRANSFER-REQUIRED 'FALSE' X-NOT-HUMAN-READABLE 'FALSE' )",
            {
                'oid': '1.2.3.4',
                'desc': 'Weirdest test äöüÄÖÜß',
                'x_subst': 'blurbBlabber',
                'not_human_readable': False,
                'x_binary_transfer_required': False,
            },
        ),
    )


class TestMatchingRule(SchemaDescriptionTest):
    """
    test ldap0.schema.models.MatchingRule
    """
    schema_class = MatchingRule
    class_attrs = (
        'oid',
        'names',
        'desc',
        'obsolete',
        'syntax',
    )
    test_cases = (
        (
            "( 1.3.6.1.1.16.3 NAME 'UUIDOrderingMatch' SYNTAX 1.3.6.1.1.16.1 )",
            {
                'oid': '1.3.6.1.1.16.3',
                'names': ('UUIDOrderingMatch', ),
                'obsolete': False,
                'syntax': '1.3.6.1.1.16.1',
            },
        ),
        (
            "( 12.34.56.78.90 NAME ( fooBarMatch $ barFooMatch ) OBSOLETE SYNTAX 1.2.3.4.5.6.7.8.9 )",
            {
                'oid': '12.34.56.78.90',
                'names': ('fooBarMatch', 'barFooMatch'),
                'obsolete': True,
                'syntax': '1.2.3.4.5.6.7.8.9',
            },
        ),
    )


class TestMatchingRuleUse(SchemaDescriptionTest):
    """
    test ldap0.schema.models.MatchingRuleUse
    """
    schema_class = MatchingRuleUse
    class_attrs = (
        'oid',
        'names',
        'desc',
        'obsolete',
        'applies',
    )
    test_cases = (
        (
            "( 1.3.6.1.4.1.4203.666.11.2.5 NAME 'CSNSIDMatch' APPLIES ( entryCSN $ namingCSN $ contextCSN ) )",
            {
                'oid': '1.3.6.1.4.1.4203.666.11.2.5',
                'names': ('CSNSIDMatch', ),
                'desc': None,
                'obsolete': False,
                'applies': ('entryCSN', 'namingCSN', 'contextCSN'),
            },
        ),
        (
            "( 1.3.6.1.4.1.4203.666.11.2.5 NAME 'CSNSIDMatch' DESC 'äöüÄÖÜß' OBSOLETE )",
            {
                'oid': '1.3.6.1.4.1.4203.666.11.2.5',
                'names': ('CSNSIDMatch', ),
                'desc': 'äöüÄÖÜß',
                'obsolete': True,
                'applies': (),
            },
        ),
    )


class TestDITContentRule(SchemaDescriptionTest):
    """
    test ldap0.schema.models.DITContentRule
    """
    schema_class = DITContentRule
    class_attrs = (
        'oid',
        'names',
        'desc',
        'obsolete',
        'aux',
        'must',
        'may',
        'nots',
    )
    test_cases = (
        (
            "( 2.16.840.1.113730.3.2.2 NAME 'inetOrgPerson-dcr' DESC 'content of inetOrgPerson entries' AUX ( pkiUser $ posixAccount $ inetLocalMailRecipient $ shadowAccount$ simpleSecurityObject ) NOT x121Address )",
            {
                'oid': '2.16.840.1.113730.3.2.2',
                'names': ('inetOrgPerson-dcr', ),
                'desc': 'content of inetOrgPerson entries',
                'obsolete': False,
                'aux': ('pkiUser', 'posixAccount', 'inetLocalMailRecipient', 'shadowAccount', 'simpleSecurityObject'),
                'must': (),
                'may': (),
                'nots': ('x121Address',),
            },
        ),
        (
            "( 1.2.3.4 NAME 'weird-oc-dcr' DESC 'äöüÄÖÜß' OBSOLETE MUST ( a $ b ) MAY ( c $ d ) )",
            {
                'oid': '1.2.3.4',
                'names': ('weird-oc-dcr', ),
                'desc': 'äöüÄÖÜß',
                'obsolete': True,
                'aux': (),
                'must': ('a', 'b'),
                'may': ('c', 'd'),
                'nots': (),
            },
        ),
    )


class TestDITStructureRule(SchemaDescriptionTest):
    """
    test ldap0.schema.models.DITStructureRule
    """
    schema_class = DITStructureRule
    class_attrs = (
        'ruleid',
        'names',
        'desc',
        'obsolete',
        'form',
        'sup',
    )
    test_cases = (
        (
            "( 104 NAME 'aeRoot-top-SR-o' FORM aeRoot-top-NF-o )",
            {
                'ruleid': '104',
                'names': ('aeRoot-top-SR-o', ),
                'desc': None,
                'obsolete': False,
                'form': 'aeRoot-top-NF-o',
                'sup': (),
            },
        ),
        (
            "( 999 NAME ( rule999 $ ruleName ) DESC 'äöüÄÖÜß' FORM 'foobar' SUP ( 1 2 3 4 ) )",
            {
                'ruleid': '999',
                'names': ('rule999', 'ruleName'),
                'desc': 'äöüÄÖÜß',
                'obsolete': False,
                'form': 'foobar',
                'sup': ('1', '2', '3', '4'),
            },
        ),
    )


class TestNameForm(SchemaDescriptionTest):
    """
    test ldap0.schema.models.NameForm
    """
    schema_class = NameForm
    class_attrs = (
        'oid',
        'names',
        'desc',
        'obsolete',
        'oc',
        'must',
        'may',
    )
    test_cases = (
        (
            "( 1.3.6.1.4.1.5427.1.389.42.15.103 NAME 'aeRoot-top-NF-o' OC aeRoot MUST ( o ) )",
            {
                'oid': '1.3.6.1.4.1.5427.1.389.42.15.103',
                'names': ('aeRoot-top-NF-o', ),
                'desc': None,
                'obsolete': False,
                'oc': 'aeRoot',
                'must': ('o',),
                'may': (),
            },
        ),
        (
            "( 1.2.3.4.5.6 NAME 'testNameForm' OC testObjectClass DESC 'äöüÄÖÜß' OBSOLETE MUST ( cn ) MAY ( mail $ uniqueIdentifier ) )",
            {
                'oid': '1.2.3.4.5.6',
                'names': ('testNameForm', ),
                'desc': 'äöüÄÖÜß',
                'obsolete': True,
                'oc': 'testObjectClass',
                'must': ('cn',),
                'may': ('mail', 'uniqueIdentifier'),
            },
        ),
    )


class SubschemaTest(unittest.TestCase):
    """
    base class for tests requiring the subschema to be loaded
    """
    schema_uri = 'file:tests/ldif/subschema-openldap-all.ldif'
    maxDiff = 10000

    @classmethod
    def setUpClass(cls):
        super(SubschemaTest, cls).setUpClass()
        cls.sub_schema = ldap0.schema.util.urlfetch(cls.schema_uri, check_uniqueness=True)[1]


class TestEntry(SubschemaTest):
    """
    test ldap0.schema.models.Entry
    """

    def test001_correct_entry(self):
        ldap_entry = {
            b'objectClass': [b'person', b'pilotPerson'],
            b'cn': [b'Michael Str\303\266der', b'Michael Stroeder'],
            b'sn': [b'Str\303\266der'],
            b'givenName': [b'Michael'],
            b'c': [b'DE'],
        }
        e1 = Entry(
            self.sub_schema,
            'cn=Michael Ströder,ou=test',
            ldap_entry,
        )
        self.assertEqual(sorted(e1.keys()), ['c', 'cn', 'givenName', 'objectClass', 'sn'])
        self.assertEqual('2.5.4.3' in e1, True)
        self.assertEqual('cn' in e1, True)
        self.assertEqual(e1['cn'], [b'Michael Str\xc3\xb6der', b'Michael Stroeder'])
        self.assertEqual(e1['cn'], e1['2.5.4.3'])
        self.assertEqual(e1['cn'], e1['commonName'])
        self.assertEqual('mail' in e1, False)
        with self.assertRaises(KeyError):
            e1['mail']
        e1.update({b'mail': [b'michael@example.com']})
        self.assertEqual('mail' in e1, True)
        self.assertEqual(e1['mail'], [b'michael@example.com'])
        self.assertEqual(e1['mail'], e1['0.9.2342.19200300.100.1.3'])
        del e1['mail']
        with self.assertRaises(KeyError):
            e1['mail']
        self.assertEqual(
            sorted(e1.object_class_oid_set()),
            ['0.9.2342.19200300.100.4.4', '2.5.6.6'],
        )
        # retrieve all required and optional attributes without applying filter
        must, may = e1.attribute_types()
        self.assertEqual(sorted(must.keys()), ['2.5.4.0', '2.5.4.3', '2.5.4.4'])
        self.assertEqual(
            sorted(may.keys()),
            [
                '0.9.2342.19200300.100.1.1', '0.9.2342.19200300.100.1.2',
                '0.9.2342.19200300.100.1.20', '0.9.2342.19200300.100.1.21',
                '0.9.2342.19200300.100.1.22', '0.9.2342.19200300.100.1.3',
                '0.9.2342.19200300.100.1.39', '0.9.2342.19200300.100.1.40',
                '0.9.2342.19200300.100.1.41', '0.9.2342.19200300.100.1.42',
                '0.9.2342.19200300.100.1.45', '0.9.2342.19200300.100.1.46',
                '0.9.2342.19200300.100.1.47', '0.9.2342.19200300.100.1.5',
                '0.9.2342.19200300.100.1.53', '0.9.2342.19200300.100.1.6',
                '0.9.2342.19200300.100.1.8',
                '2.5.4.13', '2.5.4.15', '2.5.4.20', '2.5.4.28', '2.5.4.34', '2.5.4.35'
            ],
        )
        # retrieve all required and optional attributes with SUP name
        ldap_entry = {
            b'objectClass': [b'person', b'inetOrgPerson'],
            b'cn': [b'Michael Str\303\266der', b'Michael Stroeder'],
            b'sn': [b'Str\303\266der'],
            b'givenName': [b'Michael'],
        }
        e2 = Entry(
            self.sub_schema,
            'cn=Michael Ströder,ou=test',
            ldap_entry,
        )
        must, may = e2.attribute_types(
            attr_type_filter=(
                ('sup', [('name',)]),
            ),
        )
        self.assertEqual(sorted(must.keys()), ['2.5.4.3', '2.5.4.4'])
        self.assertEqual(
            sorted(may.keys()),
            ['2.5.4.10', '2.5.4.11', '2.5.4.12', '2.5.4.42', '2.5.4.43', '2.5.4.7', '2.5.4.8'],
        )


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.schema.subentry -  subschema subentry handling
"""

from itertools import combinations

from .models import \
    AttributeType, \
    DITContentRule, \
    DITStructureRule, \
    LDAPSyntax, \
    MatchingRule, \
    MatchingRuleUse, \
    NameForm, \
    ObjectClass, \
    SchemaElementOIDSet
from ..cidict import CIDict
from ..dn import DNObj


__all__ = [
    'NameNotUnique',
    'OIDNotUnique',
    'SCHEMA_ATTR_MAPPING',
    'SCHEMA_CLASS_MAPPING',
    'SubSchema',
    'SubschemaError',
]

SCHEMA_ATTR_MAPPING = {
    AttributeType: AttributeType.schema_attribute,
    DITContentRule: DITContentRule.schema_attribute,
    DITStructureRule: DITStructureRule.schema_attribute,
    LDAPSyntax: LDAPSyntax.schema_attribute,
    MatchingRule: MatchingRule.schema_attribute,
    MatchingRuleUse: MatchingRuleUse.schema_attribute,
    NameForm: NameForm.schema_attribute,
    ObjectClass: ObjectClass.schema_attribute,
}
SCHEMA_CLASS_MAPPING = CIDict({
    AttributeType.schema_attribute: AttributeType,
    DITContentRule.schema_attribute: DITContentRule,
    DITStructureRule.schema_attribute: DITStructureRule,
    LDAPSyntax.schema_attribute: LDAPSyntax,
    MatchingRule.schema_attribute: MatchingRule,
    MatchingRuleUse.schema_attribute: MatchingRuleUse,
    NameForm.schema_attribute: NameForm,
    ObjectClass.schema_attribute: ObjectClass,
})
SCHEMA_ATTRS = list(SCHEMA_CLASS_MAPPING.keys())


class SubschemaError(ValueError):
    pass


class OIDNotUnique(SubschemaError):

    def __init__(self, desc):
        self.desc = desc

    def __str__(self):
        return 'OID not unique for %s' % (self.desc)


class NameNotUnique(SubschemaError):

    def __init__(self, desc):
        self.desc = desc

    def __str__(self):
        return 'NAME not unique for %s' % (self.desc)


class SubSchema:
    """
    Arguments:

    sub_schema_sub_entry
        Dictionary usually returned by LDAP search or the LDIF parser
        containing the sub schema sub entry

    check_uniqueness
        Defines whether uniqueness of OIDs and NAME is checked.

        0
          no check
        1
          check but add schema description with work-around
        2
          check and raise exception if non-unique OID or NAME is found

    Class attributes:

    sed
      Dictionary holding the subschema information as pre-parsed
      SchemaElement objects (do not access directly!)
    name2oid
      Dictionary holding the mapping from NAMEs to OIDs
      (do not access directly!)
    non_unique_oids
      List of OIDs used at least twice in the subschema
    non_unique_names
      List of NAMEs used at least twice in the subschema for the same schema element
    """

    def __init__(
            self,
            sub_schema_sub_entry,
            subentry_dn=None,
            check_uniqueness=True,
        ):

        # Initialize all dictionaries
        self.name2oid = {}
        self.sed = {}
        self.non_unique_oids = {}
        self.non_unique_names = {}
        for c in SCHEMA_CLASS_MAPPING.values():
            self.name2oid[c] = CIDict(keytype=str)
            self.sed[c] = {}
            self.non_unique_names[c] = CIDict(keytype=str)

        # Transform entry dict to case-insensitive dict
        e = CIDict(sub_schema_sub_entry, keytype=str)

        # Build the schema registry in dictionaries
        for attr_type in SCHEMA_ATTRS:

            for attr_value in filter(None, e.get(attr_type, [])):

                se_class = SCHEMA_CLASS_MAPPING[attr_type]
                se_instance = se_class(attr_value)
                se_id = se_instance.get_id()

                if check_uniqueness and se_id in self.sed[se_class]:
                    self.non_unique_oids[se_id] = None
                    if check_uniqueness == 1:
                        # Add to subschema by adding suffix to ID
                        suffix_counter = 1
                        new_se_id = se_id
                        while new_se_id in self.sed[se_class]:
                            new_se_id = ';'.join((se_id, str(suffix_counter)))
                            suffix_counter += 1
                        se_id = new_se_id
                    elif check_uniqueness >= 2:
                        raise OIDNotUnique(attr_value)

                # Store the schema element instance in the central registry
                self.sed[se_class][se_id] = se_instance

                if hasattr(se_instance, 'names'):
                    for name in CIDict({}.fromkeys(se_instance.names)).keys():
                        if check_uniqueness and name in self.name2oid[se_class]:
                            self.non_unique_names[se_class][se_id] = None
                            raise NameNotUnique(attr_value)
                        self.name2oid[se_class][name] = se_id

        # Turn dict into list maybe more handy for applications
        self.non_unique_oids = self.non_unique_oids.keys()
        self.subentry_dn = subentry_dn
        self.no_user_mod_attr_oids = self.determine_no_user_mod_attrs()
        # end of SubSchema()

    def ldap_entry(self):
        """
        Returns a dictionary containing the sub schema sub entry
        """
        # Initialize the dictionary with empty lists
        entry = {}
        # Collect the schema elements and store them in
        # entry's attributes
        for se_class in self.sed.keys():
            for se in self.sed[se_class].values():
                se_str = str(se)
                try:
                    entry[SCHEMA_ATTR_MAPPING[se_class]].append(se_str)
                except KeyError:
                    entry[SCHEMA_ATTR_MAPPING[se_class]] = [se_str]
        return entry
        # end of ldap_entry()

    def listall(self, schema_element_class, schema_element_filters=None):
        """
        Returns a list of OIDs of all available schema
        elements of a given schema element class.
        """
        avail_se = self.sed[schema_element_class]
        if schema_element_filters:
            result = []
            for se_key in avail_se.keys():
                se = avail_se[se_key]
                for fk, fv in schema_element_filters:
                    try:
                        if getattr(se, fk) in fv:
                            result.append(se_key)
                    except AttributeError:
                        pass
        else:
            result = avail_se.keys()
        return result
        # end of listall()

    def tree(self, schema_element_class, schema_element_filters=None):
        """
        Returns a ldap0.cidict.CIDict dictionary representing the
        tree structure of the schema elements.
        """
        assert schema_element_class in [ObjectClass, AttributeType]
        avail_se = self.listall(schema_element_class, schema_element_filters)
        top_node = '_'
        tree = CIDict({top_node:[]})
        # 1. Pass: Register all nodes
        for se in avail_se:
            tree[se] = []
        # 2. Pass: Register all sup references
        for se_oid in avail_se:
            se_obj = self.get_obj(schema_element_class, se_oid, None)
            if se_obj.__class__ != schema_element_class:
                # Ignore schema elements not matching schema_element_class.
                # This helps with falsely assigned OIDs.
                continue
            assert se_obj.__class__ == schema_element_class, TypeError(
                "Schema element referenced by %s must be of class %s, was %s" % (
                    se_oid,
                    schema_element_class.__name__,
                    se_obj.__class__
                )
            )
            for s in se_obj.sup or ('_',):
                sup_oid = self.get_oid(schema_element_class, s)
                try:
                    tree[sup_oid].append(se_oid)
                except Exception:
                    pass
        return tree
        # end of tree()

    def get_oid(self, se_class, nameoroid, raise_keyerror=0):
        """
        Get an OID by name or OID
        """
        nameoroid_stripped = nameoroid.split(';')[0].strip()
        if nameoroid_stripped in self.sed[se_class]:
            # name_or_oid is already a registered OID
            return nameoroid_stripped
        try:
            # try to look up OID by name
            result_oid = self.name2oid[se_class][nameoroid_stripped]
        except KeyError:
            if raise_keyerror:
                # raise error so zhe caller has to deal with it
                raise KeyError(
                    'No registered %s-OID for nameoroid %r' % (
                        se_class.__name__,
                        nameoroid_stripped
                    )
                )
            # just return the name
            result_oid = nameoroid_stripped
        return result_oid
        # end of get_oid()

    def get_inheritedattr(self, se_class, nameoroid, name):
        """
        Get a possibly inherited attribute specified by name
        of a schema element specified by nameoroid.
        Returns None if class attribute is not set at all.

        Raises KeyError if no schema element is found by nameoroid.
        """
        se = self.sed[se_class][self.get_oid(se_class, nameoroid)]
        try:
            result = getattr(se, name)
        except AttributeError:
            result = None
        if result is None and se.sup:
            result = self.get_inheritedattr(se_class, se.sup[0], name)
        return result
        # end of get_inheritedattr()

    def get_obj(self, se_class, nameoroid, default=None, raise_keyerror=0):
        """
        Get a schema element by name or OID
        """
        se_oid = self.get_oid(se_class, nameoroid)
        try:
            se_obj = self.sed[se_class][se_oid]
        except KeyError:
            if raise_keyerror:
                raise KeyError(
                    'No ldap0.schema.%s instance with nameoroid %r and se_oid %r' % (
                        se_class.__name__,
                        nameoroid,
                        se_oid
                    )
                )
            se_obj = default
        return se_obj
        # end of get_obj()

    def get_inheritedobj(self, se_class, nameoroid, inherited=None):
        """
        Get a schema element by name or OID with all class attributes
        set including inherited class attributes
        """
        import copy
        inherited = inherited or []
        se = copy.copy(self.sed[se_class].get(self.get_oid(se_class, nameoroid)))
        if se and hasattr(se, 'sup'):
            for class_attr_name in inherited:
                setattr(
                    se,
                    class_attr_name,
                    self.get_inheritedattr(se_class, nameoroid, class_attr_name)
                )
        return se
        # end of get_inheritedobj()

    def get_syntax(self, nameoroid):
        """
        Get the syntax of an attribute type specified by name or OID
        """
        try:
            syntax = self.get_inheritedattr(AttributeType, nameoroid, 'syntax')
        except KeyError:
            return None
        return syntax

    def get_structural_oc(self, oc_list):
        """
        Returns OID of structural object class in oc_list
        if any is present. Returns None else.
        """
        # Get tree of all STRUCTURAL object classes
        oc_tree = self.tree(ObjectClass, [('kind', {0})])
        # Filter all STRUCTURAL object classes
        struct_ocs = {}
        for oc_nameoroid in oc_list:
            oc_se = self.get_obj(ObjectClass, oc_nameoroid, None)
            if oc_se and oc_se.kind == 0:
                struct_ocs[oc_se.oid] = None
        result = None
        struct_oc_list = list(struct_ocs.keys())
        while struct_oc_list:
            oid = struct_oc_list.pop()
            for child_oid in oc_tree[oid]:
                if self.get_oid(ObjectClass, child_oid) in struct_ocs:
                    break
            else:
                result = oid
        return result
        # end of get_structural_oc()

    def get_applicable_aux_classes(self, nameoroid):
        """
        Return a list of the applicable AUXILIARY object classes
        for a STRUCTURAL object class specified by 'nameoroid'
        if the object class is governed by a DIT content rule.
        If there's no DIT content rule all available AUXILIARY
        object classes are returned.
        """
        content_rule = self.get_obj(DITContentRule, nameoroid)
        if content_rule:
            # Return AUXILIARY object classes from DITContentRule instance
            return content_rule.aux
        # list all AUXILIARY object classes
        return self.listall(ObjectClass, [('kind', [2])])
        # end of get_applicable_aux_classes()

    def attribute_types(
            self,
            object_class_list,
            attr_type_filter=None,
            raise_keyerror=1,
            ignore_dit_content_rule=0,
        ):
        """
        Returns a 2-tuple of all must and may attributes including
        all inherited attributes of superior object classes
        by walking up classes along the SUP attribute.

        The attributes are stored in a ldap0.cidict.CIDict dictionary.

        object_class_list
            list of strings specifying object class names or OIDs
        attr_type_filter
            list of 2-tuples containing lists of class attributes
            which has to be matched
        raise_keyerror
            All KeyError exceptions for non-existent schema elements
            are ignored
        ignore_dit_content_rule
            A DIT content rule governing the structural object class
            is ignored
        """

        # Map object_class_list to object_class_oids (list of OIDs)
        object_class_oids = [
            self.get_oid(ObjectClass, o)
            for o in object_class_list
        ]
        # Initialize
        oid_cache = {}

        r_must, r_may = CIDict(), CIDict()
        if '1.3.6.1.4.1.1466.101.120.111' in object_class_oids:
            # Object class 'extensibleObject' MAY carry every attribute type
            for at_obj in self.sed[AttributeType].values():
                r_may[at_obj.oid] = at_obj

        # Loop over OIDs of all given object classes
        while object_class_oids:
            object_class_oid = object_class_oids.pop(0)
            # Check whether the objectClass with this OID
            # has already been processed
            if object_class_oid in oid_cache:
                continue
            # Cache this OID as already being processed
            oid_cache[object_class_oid] = None
            try:
                object_class = self.sed[ObjectClass][object_class_oid]
            except KeyError:
                if raise_keyerror:
                    raise
                # Ignore this object class
                continue
            assert isinstance(object_class, ObjectClass)
            assert hasattr(object_class, 'must'), ValueError(object_class_oid)
            assert hasattr(object_class, 'may'), ValueError(object_class_oid)
            for a in object_class.must:
                se_oid = self.get_oid(AttributeType, a, raise_keyerror=raise_keyerror)
                r_must[se_oid] = self.get_obj(AttributeType, se_oid, raise_keyerror=raise_keyerror)
            for a in object_class.may:
                se_oid = self.get_oid(AttributeType, a, raise_keyerror=raise_keyerror)
                r_may[se_oid] = self.get_obj(AttributeType, se_oid, raise_keyerror=raise_keyerror)

            object_class_oids.extend([
                self.get_oid(ObjectClass, o)
                for o in object_class.sup
            ])

        # Process DIT content rules
        if not ignore_dit_content_rule:
            structural_oc = self.get_structural_oc(object_class_list)
            if structural_oc:
                # Process applicable DIT content rule
                try:
                    dit_content_rule = self.get_obj(DITContentRule, structural_oc, raise_keyerror=1)
                except KeyError:
                    # Not DIT content rule found for structural objectclass
                    pass
                else:
                    for a in dit_content_rule.must:
                        se_oid = self.get_oid(AttributeType, a, raise_keyerror=raise_keyerror)
                        r_must[se_oid] = self.get_obj(
                            AttributeType,
                            se_oid,
                            raise_keyerror=raise_keyerror
                        )
                    for a in dit_content_rule.may:
                        se_oid = self.get_oid(AttributeType, a, raise_keyerror=raise_keyerror)
                        r_may[se_oid] = self.get_obj(
                            AttributeType,
                            se_oid,
                            raise_keyerror=raise_keyerror
                        )
                    for a in dit_content_rule.nots:
                        a_oid = self.get_oid(AttributeType, a, raise_keyerror=raise_keyerror)
                        try:
                            del r_may[a_oid]
                        except KeyError:
                            pass

        # Remove all mandantory attribute types from
        # optional attribute type list
        for a in list(r_may.keys()):
            if a in r_must:
                del r_may[a]

        # Apply attr_type_filter to results
        if attr_type_filter:
            for l in [r_must, r_may]:
                for a in list(l.keys()):
                    for afk, afv in attr_type_filter:
                        try:
                            schema_attr_type = self.sed[AttributeType][a]
                        except KeyError:
                            if raise_keyerror:
                                raise KeyError(
                                    'Attribute type %r not found in subschema' % (a)
                                )
                            # If there's no schema element for this attribute type
                            # but still KeyError is to be ignored we filter it away
                            del l[a]
                            break
                        else:
                            if not getattr(schema_attr_type, afk) in afv:
                                del l[a]
                                break

        return r_must, r_may
        # end of attribute_types()

    def get_all_operational_attributes(self, only_user_editable=False):
        """
        Returns SchemaElementOIDSet with all operational attributes
        """
        res = SchemaElementOIDSet(self, AttributeType, [])
        for at_obj in self.sed[AttributeType].values():
            if at_obj.usage != 0 and (
                    not only_user_editable or
                    not (at_obj.no_user_mod or at_obj.collective)
                ):
                res.add(at_obj.oid)
        return res
        # end of get_all_operational_attributes()

    def determine_no_user_mod_attrs(self):
        result = {}.fromkeys([
            a.oid
            for a in self.sed[AttributeType].values()
            if a.no_user_mod
        ])
        return result
        # end of determine_no_user_mod_attrs()

    def get_associated_name_forms(self, structural_object_class_oid):
        """
        Returns a list of instances of NameForm
        representing all name forms associated with the current structural
        object class of this entry.

        The structural object class is determined by attribute
        'structuralObjectClass' if it exists or by calling
        method get_structural_oc() if not.
        """
        if structural_object_class_oid is None:
            return []
        structural_object_class_obj = self.get_obj(ObjectClass, structural_object_class_oid)
        if structural_object_class_obj:
            structural_object_class_names = [
                oc_name.lower()
                for oc_name in structural_object_class_obj.names or ()
            ]
        else:
            structural_object_class_names = ()
        result = []
        for _, name_form_obj in self.sed[NameForm].items():
            if not name_form_obj.obsolete and (
                    name_form_obj.oc == structural_object_class_oid or \
                    name_form_obj.oc.lower() in structural_object_class_names
                ):
                result.append(name_form_obj)
        return result
        # end of get_associated_name_forms()

    def get_rdn_variants(self, structural_object_class_oid):
        rdn_variants = []
        for name_form_obj in self.get_associated_name_forms(structural_object_class_oid):
            rdn_variants.append((name_form_obj, name_form_obj.must))
            for l in range(1, len(name_form_obj.may)):
                for i in combinations(name_form_obj.may, l):
                    rdn_variants.append((name_form_obj, name_form_obj.must+i))
        return rdn_variants
        # end of get_rdn_variants()

    def get_rdn_templates(self, structural_object_class_oid):
        """
        convert the tuple RDN combinations to RDN template strings
        """
        rdn_attr_tuples = {}.fromkeys([
            rdn_attr_tuple
            for name_form_obj, rdn_attr_tuple in self.get_rdn_variants(structural_object_class_oid)
        ]).keys()
        return [
            '+'.join([
                '%s=' % (attr_type)
                for attr_type in attr_types
            ])
            for attr_types in rdn_attr_tuples
        ]
        # end of get_rdn_templates()

    def get_applicable_name_form_objs(self, dn, structural_object_class_oid):
        """
        Returns a list of instances of NameForm
        representing all name form associated with the current structural
        object class of this entry and matching the current RDN.
        """
        if dn:
            rdn_list = DNObj.from_str(dn)[0]
            current_rdn_attrs = [attr_type.lower() for attr_type, _ in rdn_list]
            current_rdn_attrs.sort()
        else:
            current_rdn_attrs = []
        result = []
        for name_form_obj, rdn_attr_tuple in self.get_rdn_variants(structural_object_class_oid):
            name_form_rdn_attrs = [attr_type.lower() for attr_type in rdn_attr_tuple]
            name_form_rdn_attrs.sort()
            if current_rdn_attrs == name_form_rdn_attrs:
                result.append(name_form_obj)
        return result
        # end of get_applicable_name_form_objs()

    def get_possible_dit_structure_rules(self, dn, structural_object_class_oid):
        name_form_identifiers = CIDict()
        for name_form_obj in self.get_applicable_name_form_objs(dn, structural_object_class_oid):
            name_form_identifiers[name_form_obj.oid] = None
        dit_struct_ruleids = {}
        for dit_struct_rule_obj in self.sed[DITStructureRule].values():
            name_form_obj = self.get_obj(NameForm, dit_struct_rule_obj.form)
            if (
                    name_form_obj is not None
                    and (name_form_obj.oid in name_form_identifiers)
                    and (self.get_oid(ObjectClass, name_form_obj.oc) == structural_object_class_oid)
                ):
                dit_struct_ruleids[dit_struct_rule_obj.ruleid] = dit_struct_rule_obj
        return dit_struct_ruleids.keys()
        # end of get_possible_dit_structure_rules()

    def get_subord_structural_oc_names(self, ruleid):
        subord_structural_oc_oids = {}
        subord_structural_ruleids = {}
        for dit_struct_rule_obj in self.sed[DITStructureRule].values():
            for sup in dit_struct_rule_obj.sup:
                if sup == ruleid:
                    subord_structural_ruleids[dit_struct_rule_obj.ruleid] = None
                    name_form_obj = self.get_obj(
                        NameForm,
                        dit_struct_rule_obj.form,
                    )
                    if name_form_obj:
                        name_form_oid = self.get_oid(ObjectClass, name_form_obj.oc)
                        subord_structural_oc_oids[name_form_oid] = None
        result = []
        for oc_oid in subord_structural_oc_oids.keys():
            oc_obj = self.get_obj(ObjectClass, oc_oid)
            if oc_obj and oc_obj.names:
                result.append(oc_obj.names[0])
            else:
                result.append(oc_oid)
        return subord_structural_ruleids.keys(), result
        # end of get_subord_structural_oc_names()

    def get_superior_structural_oc_names(self, ruleid):
        try:
            dit_struct_rule_obj = self.sed[DITStructureRule][ruleid]
        except KeyError:
            return None
        else:
            result = []
            sup_ruleids = []
            for sup_ruleid in dit_struct_rule_obj.sup:
                try:
                    sup_dit_struct_rule_obj = self.sed[DITStructureRule][sup_ruleid]
                except KeyError:
                    pass
                else:
                    if sup_dit_struct_rule_obj.form:
                        sup_name_form_obj = self.get_obj(
                            NameForm,
                            sup_dit_struct_rule_obj.form,
                        )
                        if sup_name_form_obj:
                            sup_ruleids.append(sup_ruleid)
                            result.append(sup_name_form_obj.oc)
        return sup_ruleids, result
        # end of get_superior_structural_oc_names()



# -*- coding: ascii -*-
"""
ldap0.extop.dds - Classes for Dynamic Entries extended operations
(see RFC 2589)
"""

# Imports from pyasn1
from pyasn1.type import namedtype, univ, tag
from pyasn1.codec.der import encoder, decoder
from pyasn1_modules.rfc2251 import LDAPDN

from . import ExtendedRequest, ExtendedResponse


class RefreshRequest(ExtendedRequest):
    """
    Refresh Request
    (see https://tools.ietf.org/html/rfc2589#section-4.1)
    """
    __slots__ = (
        'entryName',
        'requestTtl',
    )
    requestName: str = '1.3.6.1.4.1.1466.101.119.1'
    defaultRequestTtl = 86400

    class RefreshRequestValue(univ.Sequence):
        """
        SEQUENCE {
            entryName  [0] LDAPDN,
            requestTtl [1] INTEGER
        }
        (see https://tools.ietf.org/html/rfc2589#section-4.1)
        """
        componentType = namedtype.NamedTypes(
            namedtype.NamedType(
                'entryName',
                LDAPDN().subtype(
                    implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0)
                )
            ),
            namedtype.NamedType(
                'requestTtl',
                univ.Integer().subtype(
                    implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1)
                )
            ),
        )

    def __init__(self, entryName, requestTtl=None):
        ExtendedRequest.__init__(self)
        self.entryName = entryName
        self.requestTtl = requestTtl or self.defaultRequestTtl

    def encode(self) -> bytes:
        req_value = self.RefreshRequestValue()
        req_value.setComponentByName(
            'entryName',
            LDAPDN(self.entryName.encode(self.encoding)).subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0)
            )
        )
        req_value.setComponentByName(
            'requestTtl',
            univ.Integer(self.requestTtl).subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1)
            )
        )
        return encoder.encode(req_value)


class RefreshResponse(ExtendedResponse):
    """
    Refresh Response
    (see https://tools.ietf.org/html/rfc2589#section-4.2)
    """
    __slots__ = (
        'responseTtl',
    )
    responseName = '1.3.6.1.4.1.1466.101.119.1'

    class RefreshResponseValue(univ.Sequence):
        """
        SEQUENCE {
            responseTtl [1] INTEGER
        }
        """
        componentType = namedtype.NamedTypes(
            namedtype.NamedType(
                'responseTtl',
                univ.Integer().subtype(
                    implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1)
                )
            )
        )

    def decode(self, value: bytes):
        refresh_response, _ = decoder.decode(
            value,
            asn1Spec=self.RefreshResponseValue(),
        )
        self.responseTtl = int(refresh_response.getComponentByName('responseTtl'))
        return self.responseTtl



# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.controls for extended controls and
extended operations only usable with certain slapd overlays not
available in all OpenLDAP builds
"""

import os
import unittest

from ldap0.test import SlapdTestCase

# Switch off processing .ldaprc or ldap0.conf before importing _ldap
os.environ['LDAPNOINIT'] = '1'

import ldap0
from ldap0.ldapobject import LDAPObject
from ldap0.res import SearchResultEntry
from ldap0.controls.simple import ManageDSAITControl
from ldap0.controls.deref import DereferenceControl
from ldap0.controls.openldap import SearchNoOpControl
from ldap0.controls.sss import SSSRequestControl, SSSResponseControl
from ldap0.controls.vlv import VLVRequestControl, VLVResponseControl
from ldap0.controls.simple import (
    AuthorizationIdentityRequestControl,
    AuthorizationIdentityResponseControl,
)

from ldap0.extop.dds import RefreshRequest, RefreshResponse


LDIF_TEMPLATE = """dn: %(suffix)s
objectClass: organization
o: %(o)s

dn: %(rootdn)s
objectClass: applicationProcess
objectClass: simpleSecurityObject
cn: %(rootcn)s
userPassword: %(rootpw)s

dn: cn=user1,%(suffix)s
objectClass: applicationProcess
objectClass: simpleSecurityObject
cn: user1
userPassword: user1_pw

dn: cn=Foo2,%(suffix)s
objectClass: organizationalRole
cn: Foo2

dn: cn=Foo1,%(suffix)s
objectClass: organizationalRole
cn: Foo1

dn: cn=Foo3,%(suffix)s
objectClass: organizationalRole
cn: Foo3
roleOccupant: cn=user1,%(suffix)s

dn: ou=Container,%(suffix)s
objectClass: organizationalUnit
ou: Container

dn: cn=Foo5,ou=Container,%(suffix)s
objectClass: organizationalRole
cn: Foo5
roleOccupant: cn=user1,%(suffix)s

dn: cn=Foo4,ou=Container,%(suffix)s
objectClass: organizationalRole
cn: Foo4

dn: cn=Foo7,ou=Container,%(suffix)s
objectClass: organizationalRole
cn: Foo7
roleOccupant: cn=user1,%(suffix)s
roleOccupant: cn=Foo1,%(suffix)s
roleOccupant: cn=Foo6,ou=Container,%(suffix)s

dn: cn=Foo6,ou=Container,%(suffix)s
objectClass: organizationalRole
cn: Foo6

dn: dc=ref1,%(suffix)s
objectClass: referral
objectClass: dcObject
dc: ref1
ref: ldap://ldap.example.com/dc=example,dc=com??one?(objectClass=*)

dn: dc=ref2,%(suffix)s
objectClass: referral
objectClass: dcObject
dc: ref2
ref: ldap://ldap.example.net/

dn: ou=Groups,%(suffix)s
objectClass: organizationalUnit
ou: Container

dn: cn=Bar1,ou=Groups,%(suffix)s
objectClass: posixGroup
cn: Bar1
gidNumber: 10001

dn: cn=Bar3,ou=Groups,%(suffix)s
objectClass: posixGroup
cn: Bar3
gidNumber: 10003

dn: cn=Bar4,ou=Groups,%(suffix)s
objectClass: posixGroup
cn: Bar4
gidNumber: 10004

dn: cn=Bar2,ou=Groups,%(suffix)s
objectClass: posixGroup
cn: Bar2
gidNumber: 10002

"""


class ExtensionTestCase(SlapdTestCase):
    """
    test LDAP search operations
    """
    ldap_object_class = LDAPObject
    maxDiff = None
    trace_level = 0

    @classmethod
    def setUpClass(cls):
        super(ExtensionTestCase, cls).setUpClass()
        # insert some Foo* objects via ldapadd
        cls.server.ldapadd(
            (LDIF_TEMPLATE % {
                'suffix':cls.server.suffix,
                'rootdn':cls.server.root_dn,
                'rootcn':cls.server.root_cn,
                'rootpw':cls.server.root_pw,
                'o': cls.server.suffix.split(',')[0][3:],
            }).encode('utf-8')
        )


class TestDeRef(ExtensionTestCase):

    def test001_deref_search_s(self):
        self.enable_overlay({
            'olcOverlay': ['deref'],
            'objectClass': ['olcOverlayConfig'],
        })
        ldap_result = self._ldap_conn.search_s(
            'cn=Foo7,ou=Container,%s' % (self.server.suffix),
            ldap0.SCOPE_BASE,
            req_ctrls=[
                DereferenceControl(
                    True,
                    {'roleOccupant': ['cn', 'uid']},
                )
            ],
            cache_ttl=1.0,
        )

        self.assertEqual(
            ldap_result[0].dn_s,
            'cn=Foo7,ou=Container,%s' % (self.server.suffix)
        )
        self.assertEqual(
            ldap_result[0].entry_as,
            {
                'cn': [b'Foo7'],
                'objectClass': [b'organizationalRole'],
                'roleOccupant': [
                    b'cn=user1,%s' % (self.server.suffix.encode('utf-8')),
                    b'cn=Foo1,%s' % (self.server.suffix.encode('utf-8')),
                    b'cn=Foo6,ou=Container,%s' % (self.server.suffix.encode('utf-8'))
                ],
            },
        )
        deref_ctrl = ldap_result[0].ctrls[0]
        self.assertEqual(isinstance(deref_ctrl, DereferenceControl), True)
        self.assertEqual(
            deref_ctrl.derefRes,
            {
                'roleOccupant': [
                    SearchResultEntry(b'cn=user1,%s' % (self.server.suffix.encode('utf-8')), {b'cn': [b'user1']}),
                    SearchResultEntry(b'cn=Foo1,%s' % (self.server.suffix.encode('utf-8')), {b'cn': [b'Foo1']}),
                    SearchResultEntry(b'cn=Foo6,ou=Container,%s' % (self.server.suffix.encode('utf-8')), {b'cn': [b'Foo6']}),
                ]
            }
        )
        del ldap_result; del deref_ctrl
        ldap_result = self._ldap_conn.find_unique_entry(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
            '(cn=Foo7)',
            req_ctrls=[
                DereferenceControl(
                    True,
                    {'roleOccupant': ['cn', 'uid']},
                )
            ],
        )
        self.assertEqual(ldap_result.dn_s, 'cn=Foo7,ou=Container,%s' % (self.server.suffix))
        self.assertEqual(
            ldap_result.entry_as,
            {
                'cn': [b'Foo7'],
                'objectClass': [b'organizationalRole'],
                'roleOccupant': [
                    b'cn=user1,%s' % (self.server.suffix.encode('utf-8')),
                    b'cn=Foo1,%s' % (self.server.suffix.encode('utf-8')),
                    b'cn=Foo6,ou=Container,%s' % (self.server.suffix.encode('utf-8'))
                ],
            },
        )
        deref_ctrl = ldap_result.ctrls[0]
        self.assertEqual(isinstance(deref_ctrl, DereferenceControl), True)
        self.assertEqual(
            deref_ctrl.derefRes,
            {
                'roleOccupant': [
                    SearchResultEntry(b'cn=user1,%s' % (self.server.suffix.encode('utf-8')), {b'cn': [b'user1']}),
                    SearchResultEntry(b'cn=Foo1,%s' % (self.server.suffix.encode('utf-8')), {b'cn': [b'Foo1']}),
                    SearchResultEntry(b'cn=Foo6,ou=Container,%s' % (self.server.suffix.encode('utf-8')), {b'cn': [b'Foo6']}),
                ]
            }
        )


class TestNoopSrch(ExtensionTestCase):

    def test_noopsrch(self):
        self.enable_overlay({
            'olcOverlay': ['noopsrch'],
            'objectClass': ['olcOverlayConfig'],
        })
        expected_num_refs = LDIF_TEMPLATE.count('objectClass: referral')
        expected_num_entries = LDIF_TEMPLATE.count('dn: ') - expected_num_refs
        num_entries, num_refs = self._ldap_conn.noop_search(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
        )
        self.assertEqual(num_entries, expected_num_entries)
        self.assertEqual(num_refs, expected_num_refs)
        num_entries, num_refs = self._ldap_conn.noop_search(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
            '(objectClass=organizationalRole)',
        )
        self.assertEqual(num_entries, LDIF_TEMPLATE.count('objectClass: organizationalRole'))
        self.assertEqual(num_refs, LDIF_TEMPLATE.count('objectClass: referral'))
        num_entries, num_refs = self._ldap_conn.noop_search(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
            req_ctrls=[ManageDSAITControl()]
        )
        self.assertEqual(num_entries, LDIF_TEMPLATE.count('dn: '))
        self.assertEqual(num_refs, 0)


class TestDDS(ExtensionTestCase):

    def test_dds(self):
        self.enable_overlay({
            'olcOverlay': ['dds'],
            'objectClass': ['olcOverlayConfig', 'olcDDSConfig'],
            'olcDDSdefaultTtl': ['1h'],
            'olcDDSmaxTtl': ['1d'],
            'olcDDSminTtl': ['1m'],
            'olcDDSstate': ['TRUE'],
        })
        root_dse = self._ldap_conn.read_rootdse_s(attrlist=['dynamicSubtrees'])
        if 'dynamicSubtrees' not in root_dse.entry_s:
            self.skipTest('slapo-dds not enabled in rootDSE')
        self.assertEqual(root_dse.entry_s['dynamicSubtrees'][0], self.server.suffix)
        dds_entry_dn = 'cn=dds-1,'+self.server.suffix
        self._ldap_conn.add_s(
            dds_entry_dn,
            {
                'objectClass': [b'applicationProcess', b'dynamicObject'],
                'cn': [b'dds-1'],
            },
        )
        refresh_response = self._ldap_conn.extop_s(
            RefreshRequest(
                entryName=dds_entry_dn,
                requestTtl=247,
            ),
            extop_resp_class=RefreshResponse,
        )
        self.assertEqual(refresh_response.responseTtl, 247)
        self.assertEqual(self._ldap_conn.refresh_s(dds_entry_dn, ttl=178), 178)
        self.assertEqual(self._ldap_conn.refresh_s(dds_entry_dn, ttl=42), 60)
        self.assertEqual(self._ldap_conn.refresh_s(dds_entry_dn), 86400)


class TestSssVlv(ExtensionTestCase):

    def test_sssvlv(self):
        self.enable_overlay(
            {
                'olcOverlay': ['sssvlv'],
                'objectClass': ['olcOverlayConfig', 'olcSssVlvConfig'],
                'olcSssVlvMax': ['0'],
                'olcSssVlvMaxKeys': ['5'],
                'olcSssVlvMaxPerConn': ['5'],
            },
        )
        wanted_results = [
            (
                'cn=Bar%d,ou=Groups,%s' % (i, self.server.suffix),
                {'cn': [b'Bar%d' % i]},
            )
            for i in range(4, 0, -1)
        ]

        for ordr in (
            ('-cn:caseExactOrderingMatch',),
            ('-cn:caseExactOrderingMatch',),
            ('-gidNumber:integerOrderingMatch',),
            ('-gidNumber',),
            ('-gidNumber', 'createTimestamp',),
        ):
            ldap_result = self._ldap_conn.search_s(
                self.server.suffix,
                ldap0.SCOPE_SUBTREE,
                '(objectClass=posixGroup)',
                attrlist=['cn'],
                req_ctrls=[SSSRequestControl(criticality=True, ordering_rules=ordr)],
            )
            self.assertEqual([(lr.dn_s, lr.entry_as) for lr in ldap_result], wanted_results)
        msg_id = self._ldap_conn.search(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
            '(objectClass:=organizationalRole)',
            attrlist=['cn'],
            req_ctrls=[
                SSSRequestControl(
                    criticality=True,
                    ordering_rules=['cn:caseExactOrderingMatch']
                )
            ],
            sizelimit=1,
        )
        ldap_result = []
        try:
            for ldap_res in self._ldap_conn.results(msg_id):
                ldap_result.extend(ldap_res.rdata)
        except ldap0.SIZELIMIT_EXCEEDED:
            pass
        self.assertEqual(
            [(lr.dn_s, lr.entry_as) for lr in ldap_result],
            [
                (
                    'cn=Foo1,%s' % (self.server.suffix),
                    {'cn': [b'Foo1']}
                ),
            ]
        )


class TestAuthzID(ExtensionTestCase):

    def test001_authz_id(self):
        root_dse = self._ldap_conn.read_rootdse_s(attrlist=['supportedControl'])
        self.assertNotIn(AuthorizationIdentityRequestControl.controlType, root_dse.entry_s['supportedControl'])
        self.enable_overlay(
            {
                'olcOverlay': ['authzid'],
                'objectClass': ['olcOverlayConfig'],
            },
            config_dn='olcDatabase={-1}frontend,cn=config',
        )
        root_dse = self._ldap_conn.read_rootdse_s(attrlist=['supportedControl'])
        self.assertIn(AuthorizationIdentityRequestControl.controlType, root_dse.entry_s['supportedControl'])
        bind_dn = 'cn=user1,'+self.server.suffix
        bind_pw = 'user1_pw'.encode('ascii')
        with self.ldap_object_class(self.server.ldap_uri) as l:
            bind_res = l.simple_bind_s(bind_dn, bind_pw)
            whoami_dn = l._whoami_dn
        self.assertEqual(whoami_dn, None)
        with self.ldap_object_class(self.server.ldap_uri) as l:
            bind_res = l.simple_bind_s(
                bind_dn,
                bind_pw,
                req_ctrls=[AuthorizationIdentityRequestControl()],
            )
            whoami_dn = l._whoami_dn
        self.assertEqual(len(bind_res.ctrls), 1)
        self.assertIsInstance(bind_res.ctrls[0], AuthorizationIdentityResponseControl)
        self.assertEqual(
            bind_res.ctrls[0].authzId.decode('utf-8'),
            'dn:'+bind_dn
        )
        self.assertEqual(whoami_dn, bind_dn)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
test consistency of meta information
"""

# from Python's standard lib
import unittest

import ldap0.__about__

class TestVersionInfo(unittest.TestCase):

    def test_version(self):
        self.assertEqual(
            ldap0.__about__.__version_info__.major, ldap0.__about__.__version_info__[0],
        )
        self.assertEqual(
            ldap0.__about__.__version_info__.minor, ldap0.__about__.__version_info__[1],
        )
        self.assertEqual(
            ldap0.__about__.__version_info__.micro, ldap0.__about__.__version_info__[2],
        )
        self.assertEqual(
            tuple(int(val) for val in ldap0.__about__.__version__.split('.', 3)),
            ldap0.__about__.__version_info__,
        )


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.extop - support classes for LDAPv3 extended operations
"""

from typing import Dict, Union, Optional


class ExtendedRequest:
    """
    Generic base class for a LDAPv3 extended operation request

    requestName
        OID as string of the LDAPv3 extended operation request
    requestValue
        value of the LDAPv3 extended operation request
        (here it is the BER-encoded ASN.1 request value)
    """
    __slots__ = (
        'requestName',
        'requestValue',
    )
    encoding: str = 'utf-8'

    defaultIntermediateResponse: Optional['IntermediateResponse'] = None

    def __init__(self, requestName=None, requestValue=None):
        if requestName is not None:
            self.requestName = requestName
        self.requestValue = requestValue

    def __repr__(self):
        return '%s(requestName=%r, requestValue=%r)' % (
            self.__class__.__name__,
            self.requestName,
            self.requestValue,
        )

    def encode(self) -> bytes:
        """
        returns the BER-encoded ASN.1 request value composed by class attributes
        set before
        """
        return self.requestValue


class ExtendedResponse:
    """
    Generic base class for a LDAPv3 extended operation response

    requestName
        OID as string of the LDAPv3 extended operation response
    encodedResponseValue
        BER-encoded ASN.1 value of the LDAPv3 extended operation response
    """
    __slots__ = (
        'responseValue',
    )
    encoding = 'utf-8'

    responseName = None

    def __init__(self, encodedResponseValue=None):
        self.responseValue = encodedResponseValue
        if encodedResponseValue is not None:
            self.decode(encodedResponseValue)

    def __repr__(self):
        return '%s(encodedResponseValue=%r)' % (
            self.__class__.__name__,
            self.responseValue,
        )

    @classmethod
    def check_resp_name(cls, name):
        """
        returns True if :name: is the correct expected responseName
        """
        return name == cls.responseName

    def decode(self, value: bytes):
        """
        decodes the BER-encoded ASN.1 extended operation response value and
        sets the appropriate class attributes
        """
        self.responseValue = value



# -*- coding: ascii -*-
"""
ldap0.base - basic classes and functions
"""

from typing import Sequence

from .typehints import BytesList, StrList, EntryBytes, EntryStr

__all__ = [
    'decode_list',
    'decode_entry_dict',
    'encode_list',
    'encode_entry_dict',
]


def decode_list(lst: Sequence[bytes], encoding: str = 'utf-8') -> StrList:
    """
    decode a sequence containing only byte-strings with given encoding
    and return a list of strings
    """
    return [item.decode(encoding) for item in lst]


def decode_entry_dict(entry: EntryBytes, encoding: str = 'utf-8') -> EntryStr:
    """
    decode an LDAP entry dict containing only
    byte-strings in keys and values with given encoding
    """
    return {
        at.decode(encoding): decode_list(avs, encoding=encoding)
        for at, avs in entry.items()
    }


def encode_list(lst: Sequence[str], encoding: str = 'utf-8') -> BytesList:
    """
    encode a sequence containing only Unicode-strings with given encoding
    and return a list of bytes
    """
    return [item.encode(encoding) for item in lst]


def encode_entry_dict(entry: EntryStr, encoding: str = 'utf-8') -> EntryBytes:
    """
    encode an LDAP entry dict containing only
    Unicode-strings in keys and values with given encoding
    """
    return {
        at.encode(encoding): encode_list(avs, encoding=encoding)
        for at, avs in entry.items()
    }



# -*- coding: ascii -*-
"""
ldap0.extop.cancel - Classes for Cancel Operation
(see RFC 3909)
"""

# Imports from pyasn1
from pyasn1.type import namedtype, univ
from pyasn1.codec.der import encoder
from pyasn1_modules.rfc2251 import MessageID

from . import ExtendedRequest, ExtendedResponse


class CancelRequest(ExtendedRequest):
    """
    Cancel Operation Request
    """
    __slots__ = (
        'cancelID',
    )
    requestName: str = '1.3.6.1.1.8'
    cancelID: int

    class CancelRequestValue(univ.Sequence):
        """
        Cancel request value
        """
        componentType = namedtype.NamedTypes(
            namedtype.NamedType(
                'cancelID',
                MessageID()
            ),
        )

    def __init__(self, cancelID: int):
        ExtendedRequest.__init__(self)
        self.cancelID = cancelID

    def encode(self) -> bytes:
        req_value = self.CancelRequestValue()
        req_value.setComponentByName(
            'cancelID',
            str(self.cancelID).encode('ascii')
        )
        return encoder.encode(req_value)


class CancelResponse(ExtendedResponse):
    """
    Cancel Operation Response (only dummy)
    """



# -*- coding: utf-8 -*-
"""
Automatic tests for C wrapper module _libldap0
"""

import os
import logging
import unittest
from errno import ENOTCONN

# Switch off processing .ldaprc or ldap0.conf before importing _ldap
os.environ['LDAPNOINIT'] = '1'

import _libldap0

from ldap0.test import SlapdTestCase
from ldap0.sasl import SaslNoninteractiveAuth
from ldap0.functions import escape_format
from ldap0.filter import escape_str as escape_filter_str


class TestLdapCExtension(SlapdTestCase):
    """
    These tests apply only to the _ldap module and therefore bypass the
    LDAPObject wrapper completely.
    """

    timeout = 5

    @classmethod
    def setUpClass(cls):
        super(TestLdapCExtension, cls).setUpClass()
        # add two initial objects after server was started and is still empty
        suffix_dc = cls.server.suffix.split(',')[0][3:]
        logging.debug(
            'adding %s and %s',
            cls.server.suffix,
            cls.server.root_dn,
        )
        cls.server.ldapadd(
            '\n'.join([
                'dn: '+cls.server.suffix,
                'objectClass: organization',
                'o:'+suffix_dc,
                '',
                'dn: '+cls.server.root_dn,
                'objectClass: applicationProcess',
                'cn: '+cls.server.root_cn,
                '',
            ]).encode('utf-8')
        )

    def setUp(self):
        try:
            self._ldap_conn
        except AttributeError:
            # open local LDAP connection
            self._ldap_conn = self._open_ldap_conn()
            msg_id = self._ldap_conn.search_ext(
                'cn=config'.encode('utf-8'),
                _libldap0.SCOPE_SUBTREE,
                escape_format(escape_filter_str, '(olcSuffix={0})', self.server.suffix).encode('utf-8'),
            )
            ldap_res = self._ldap_conn.result(msg_id)
            logging.debug('ldap_res = %r', ldap_res)

    def _open_ldap_conn(self, bind=True):
        """
        Starts a server, and returns a LDAPObject bound to it
        """
        ldap_conn = _libldap0._initialize(self.server.ldap_uri.encode('ascii'))
        ldap_conn.set_option(_libldap0.OPT_PROTOCOL_VERSION, _libldap0.VERSION3)
        if bind:
            # Perform a simple bind
            msg_id = ldap_conn.simple_bind(
                self.server.root_dn.encode('utf-8'),
                self.server.root_pw.encode('utf-8')
            )
            result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ONE, self.timeout)
            self.assertEqual(result, _libldap0.RES_BIND)
            self.assertEqual(type(msgid), type(0))
        return ldap_conn

    # Test for the existence of a whole bunch of constants
    # that the C module is supposed to export
    def test001_constants(self):
        """
        Test whether all libldap-derived constants are correct
        """
        self.assertEqual(_libldap0.PORT, 389)
        self.assertEqual(_libldap0.VERSION1, 1)
        self.assertEqual(_libldap0.VERSION2, 2)
        self.assertEqual(_libldap0.VERSION3, 3)
        # constants for result()
        self.assertEqual(_libldap0.RES_BIND, 0x61)
        self.assertEqual(_libldap0.RES_SEARCH_ENTRY, 0x64)
        self.assertEqual(_libldap0.RES_SEARCH_RESULT, 0x65)
        self.assertEqual(_libldap0.RES_MODIFY, 0x67)
        self.assertEqual(_libldap0.RES_ADD, 0x69)
        self.assertEqual(_libldap0.RES_DELETE, 0x6b)
        self.assertEqual(_libldap0.RES_MODRDN, 0x6d)
        self.assertEqual(_libldap0.RES_COMPARE, 0x6f)
        self.assertEqual(_libldap0.RES_SEARCH_REFERENCE, 0x73) # v3
        self.assertEqual(_libldap0.RES_EXTENDED, 0x78)         # v3
        #self.assertEqual(_libldap0.RES_INTERMEDIATE, 0x79)     # v3
        self.assertIsNotNone(_libldap0.RES_ANY)
        self.assertIsNotNone(_libldap0.RES_UNSOLICITED)
        self.assertIsNotNone(_libldap0.SCOPE_BASE)
        self.assertIsNotNone(_libldap0.SCOPE_ONELEVEL)
        self.assertIsNotNone(_libldap0.SCOPE_SUBTREE)
        self.assertIsNotNone(_libldap0.MOD_ADD)
        self.assertIsNotNone(_libldap0.MOD_DELETE)
        self.assertIsNotNone(_libldap0.MOD_REPLACE)
        self.assertIsNotNone(_libldap0.MOD_INCREMENT)
        # for result()
        self.assertIsNotNone(_libldap0.MSG_ONE)
        self.assertIsNotNone(_libldap0.MSG_ALL)
        self.assertIsNotNone(_libldap0.MSG_RECEIVED)
        # for OPT_DEFEF
        self.assertIsNotNone(_libldap0.DEREF_NEVER)
        self.assertIsNotNone(_libldap0.DEREF_SEARCHING)
        self.assertIsNotNone(_libldap0.DEREF_FINDING)
        self.assertIsNotNone(_libldap0.DEREF_ALWAYS)
        # for OPT_SIZELIMIT, OPT_TIMELIMIT
        self.assertIsNotNone(_libldap0.NO_LIMIT)
        # standard options
        self.assertIsNotNone(_libldap0.OPT_API_INFO)
        self.assertIsNotNone(_libldap0.OPT_DEREF)
        self.assertIsNotNone(_libldap0.OPT_SIZELIMIT)
        self.assertIsNotNone(_libldap0.OPT_TIMELIMIT)
        self.assertIsNotNone(_libldap0.OPT_REFERRALS)
        self.assertIsNotNone(_libldap0.OPT_RESTART)
        self.assertIsNotNone(_libldap0.OPT_PROTOCOL_VERSION)
        self.assertIsNotNone(_libldap0.OPT_SERVER_CONTROLS)
        self.assertIsNotNone(_libldap0.OPT_API_FEATURE_INFO)
        self.assertIsNotNone(_libldap0.OPT_HOST_NAME)
        self.assertIsNotNone(_libldap0.OPT_ERROR_NUMBER)   # = OPT_RESULT_CODE
        self.assertIsNotNone(_libldap0.OPT_ERROR_STRING)   # = OPT_DIAGNOSITIC_MESSAGE
        self.assertIsNotNone(_libldap0.OPT_MATCHED_DN)
        # OpenLDAP specific
        self.assertIsNotNone(_libldap0.OPT_DEBUG_LEVEL)
        self.assertIsNotNone(_libldap0.OPT_TIMEOUT)
        self.assertIsNotNone(_libldap0.OPT_REFHOPLIMIT)
        self.assertIsNotNone(_libldap0.OPT_NETWORK_TIMEOUT)
        self.assertIsNotNone(_libldap0.OPT_URI)
        #self.assertIsNotNone(_libldap0.OPT_REFERRAL_URLS)
        #self.assertIsNotNone(_libldap0.OPT_SOCKBUF)
        #self.assertIsNotNone(_libldap0.OPT_DEFBASE)
        #self.assertIsNotNone(_libldap0.OPT_CONNECT_ASYNC)
        # str2dn()
        self.assertIsNotNone(_libldap0.DN_FORMAT_LDAP)
        self.assertIsNotNone(_libldap0.DN_FORMAT_LDAPV3)
        self.assertIsNotNone(_libldap0.DN_FORMAT_LDAPV2)
        self.assertIsNotNone(_libldap0.DN_FORMAT_DCE)
        self.assertIsNotNone(_libldap0.DN_FORMAT_UFN)
        self.assertIsNotNone(_libldap0.DN_FORMAT_AD_CANONICAL)
        self.assertIsNotNone(_libldap0.DN_FORMAT_MASK)
        self.assertIsNotNone(_libldap0.DN_PRETTY)
        self.assertIsNotNone(_libldap0.DN_SKIP)
        self.assertIsNotNone(_libldap0.DN_P_NOLEADTRAILSPACES)
        self.assertIsNotNone(_libldap0.DN_P_NOSPACEAFTERRDN)
        self.assertIsNotNone(_libldap0.DN_PEDANTIC)
        self.assertIsNotNone(_libldap0.AVA_NULL)
        self.assertIsNotNone(_libldap0.AVA_STRING)
        self.assertIsNotNone(_libldap0.AVA_BINARY)
        self.assertIsNotNone(_libldap0.AVA_NONPRINTABLE)

    def test002_simple_bind(self):
        ldap_conn = self._open_ldap_conn()

    def test003_simple_anonymous_bind(self):
        ldap_conn = self._open_ldap_conn(bind=False)
        msg_id = ldap_conn.simple_bind(b'', b'')
        self.assertIsInstance(msg_id, int)
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        self.assertEqual(result, _libldap0.RES_BIND)
        self.assertEqual(msgid, msg_id)
        self.assertEqual(pmsg, [])
        self.assertEqual(ctrls, [])

    def test004_anon_rootdse_search(self):
        ldap_conn = self._open_ldap_conn(bind=False)
        # see if we can get the rootdse with anon search (without prior bind)
        msg_id = ldap_conn.search_ext(
            b'',
            _libldap0.SCOPE_BASE,
            b'(objectClass=*)',
            [b'objectClass', b'namingContexts'],
        )
        self.assertIsInstance(msg_id, int)
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        self.assertEqual(result, _libldap0.RES_SEARCH_RESULT)
        self.assertEqual(pmsg[0][0], b'') # rootDSE has no dn
        self.assertEqual(msgid, msg_id)
        self.assertEqual(ctrls, [])
        root_dse = pmsg[0][1]
        self.assertTrue(b'objectClass' in root_dse)
        self.assertIn(b'OpenLDAProotDSE', root_dse[b'objectClass'])
        self.assertTrue(b'namingContexts' in root_dse)
        self.assertEqual(
            root_dse[b'namingContexts'],
            [self.server.suffix.encode('utf-8')]
        )

    def test005_unbind(self):
        ldap_conn = self._open_ldap_conn()
        msg_id = ldap_conn.unbind_ext()
        self.assertIsNone(msg_id)
        # Second attempt to unbind should yield an exception
        with self.assertRaises(_libldap0.LDAPError):
            ldap_conn.unbind_ext()

    def test006_search_ext_individual(self):
        ldap_conn = self._open_ldap_conn()
        # send search request
        msg_id = ldap_conn.search_ext(
            self.server.suffix.encode('utf-8'),
            _libldap0.SCOPE_SUBTREE,
            b'(objectClass=organization)'
        )
        self.assertIsInstance(msg_id, int)
        ldap_res = ldap_conn.result(
            msg_id,
            _libldap0.MSG_ONE,
            self.timeout
        )
        result, pmsg, msgid, ctrls = ldap_res
        # Expect to get just one object
        self.assertEqual(result, _libldap0.RES_SEARCH_ENTRY)
        self.assertEqual(len(pmsg), 1)
        self.assertEqual(len(pmsg[0]), 2)
        self.assertEqual(pmsg[0][0], self.server.suffix.encode('utf-8'))
        self.assertEqual(pmsg[0][0], self.server.suffix.encode('utf-8'))
        self.assertIn(b'organization', pmsg[0][1][b'objectClass'])
        self.assertEqual(msgid, msg_id)
        self.assertEqual(ctrls, [])

        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ONE, self.timeout)
        self.assertEqual(result, _libldap0.RES_SEARCH_RESULT)
        self.assertEqual(pmsg, [])
        self.assertEqual(msgid, msg_id)
        self.assertEqual(ctrls, [])

    def test007_abandon(self):
        ldap_conn = self._open_ldap_conn()
        msg_id = ldap_conn.search_ext(self.server.suffix.encode('utf-8'), _libldap0.SCOPE_SUBTREE, b'(objectClass=*)')
        ret = ldap_conn.abandon_ext(msg_id)
        self.assertIsNone(ret)
        with self.assertRaises(_libldap0.TIMEOUT):
            r = ldap_conn.result(msg_id, _libldap0.MSG_ALL, 0.3)  # (timeout /could/ be longer)

    def test008_search_ext_all(self):
        ldap_conn = self._open_ldap_conn()
        # send search request
        msg_id = ldap_conn.search_ext(
            self.server.suffix.encode('utf-8'),
            _libldap0.SCOPE_SUBTREE,
            b'(objectClass=*)'
        )
        self.assertIsInstance(msg_id, int)
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        # Expect to get some objects
        self.assertEqual(result, _libldap0.RES_SEARCH_RESULT)
        self.assertTrue(len(pmsg) >= 2)
        self.assertEqual(msgid, msg_id)
        self.assertEqual(ctrls, [])

    def test009_add(self):
        """
        test add operation
        """
        ldap_conn = self._open_ldap_conn()
        msg_id = ldap_conn.add_ext(
            ('cn=Foo,'+self.server.suffix).encode('utf-8'),
            [
                (b'objectClass', [b'organizationalRole']),
                (b'cn', [b'Foo']),
                (b'description', [b'testing']),
            ]
        )
        self.assertIsInstance(msg_id, int)
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        self.assertEqual(result, _libldap0.RES_ADD)
        self.assertEqual(pmsg, [])
        self.assertEqual(msgid, msg_id)
        self.assertEqual(ctrls, [])
        # search for it back
        msg_id = ldap_conn.search_ext(self.server.suffix.encode('utf-8'), _libldap0.SCOPE_SUBTREE, b'(cn=Foo)')
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        # Expect to get the objects
        self.assertEqual(result, _libldap0.RES_SEARCH_RESULT)
        self.assertEqual(len(pmsg), 1)
        self.assertEqual(msgid, msg_id)
        self.assertEqual(ctrls, [])
        self.assertEqual(
            pmsg[0],
            (
                ('cn=Foo,'+self.server.suffix).encode('utf-8'),
                {
                    b'objectClass': [b'organizationalRole'],
                    b'cn': [b'Foo'],
                    b'description': [b'testing'],
                }
            )
        )

    def test010_compare(self):
        """
        test compare operation
        """
        ldap_conn = self._open_ldap_conn()
        # first, add an object with a field we can compare on
        dn = 'cn=CompareTest,' + self.server.suffix
        msg_id = ldap_conn.add_ext(
            dn.encode('utf-8'),
            [
                (b'objectClass', [b'person']),
                (b'sn', [b'CompareTest']),
                (b'cn', [b'CompareTest']),
                (b'userPassword', [b'the_password']),
            ],
        )
        self.assertIsInstance(msg_id, int)
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        self.assertEqual(result, _libldap0.RES_ADD)
        # try a false compare
        msg_id = ldap_conn.compare_ext(dn.encode('utf-8'), b'userPassword', b'bad_string')
        with self.assertRaises(_libldap0.COMPARE_FALSE) as e:
            r = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)

        self.assertEqual(e.exception.args[0]['msgid'], msg_id)
        self.assertEqual(e.exception.args[0]['msgtype'], _libldap0.RES_COMPARE)
        self.assertEqual(e.exception.args[0]['result'], 5)
        self.assertFalse(e.exception.args[0]['ctrls'])

        # try a true compare
        msg_id = ldap_conn.compare_ext(dn.encode('utf-8'), b'userPassword', b'the_password')
        with self.assertRaises(_libldap0.COMPARE_TRUE) as e:
            r = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)

        self.assertEqual(e.exception.args[0]['msgid'], msg_id)
        self.assertEqual(e.exception.args[0]['msgtype'], _libldap0.RES_COMPARE)
        self.assertEqual(e.exception.args[0]['result'], 6)
        self.assertFalse(e.exception.args[0]['ctrls'])

        # try a compare on bad attribute
        msg_id = ldap_conn.compare_ext(dn.encode('utf-8'), b'badAttribute', b'ignoreme')
        with self.assertRaises(_libldap0.error) as e:
            r = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)

        self.assertEqual(e.exception.args[0]['msgid'], msg_id)
        self.assertEqual(e.exception.args[0]['msgtype'], _libldap0.RES_COMPARE)
        self.assertEqual(e.exception.args[0]['result'], 17)
        self.assertFalse(e.exception.args[0]['ctrls'])

    def test011_delete_no_such_object(self):
        """
        try deleting an object that doesn't exist
        """
        ldap_conn = self._open_ldap_conn()
        msg_id = ldap_conn.delete_ext(('cn=DoesNotExist,'+self.server.suffix).encode('utf-8'))
        with self.assertRaises(_libldap0.NO_SUCH_OBJECT):
            r = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)

    def test012_delete(self):
        ldap_conn = self._open_ldap_conn()
        # first, add an object we will delete
        dn = 'cn=Deleteme,'+self.server.suffix
        msg_id = ldap_conn.add_ext(
            dn.encode('utf-8'),
            [
                (b'objectClass', [b'organizationalRole']),
                (b'cn', [b'Deleteme']),
            ]
        )
        self.assertIsInstance(msg_id, int)
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        self.assertEqual(result, _libldap0.RES_ADD)
        msg_id = ldap_conn.delete_ext(dn.encode('utf-8'))
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        self.assertEqual(result, _libldap0.RES_DELETE)
        self.assertEqual(msgid, msg_id)
        self.assertEqual(pmsg, [])
        self.assertEqual(ctrls, [])

    def test013_modify_no_such_object(self):
        ldap_conn = self._open_ldap_conn()
        # try to modify an object that doesn't exist
        msg_id = ldap_conn.modify_ext(
            ('cn=DoesNotExist,'+self.server.suffix).encode('utf-8'),
            [
                (_libldap0.MOD_ADD, b'description', [b'blah']),
            ]
        )
        with self.assertRaises(_libldap0.NO_SUCH_OBJECT):
            r = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)

    def test014_modify_no_such_object_empty_attrs(self):
        """
        try deleting an object that doesn't exist
        """
        ldap_conn = self._open_ldap_conn()
        msg_id = ldap_conn.modify_ext(
            ('cn=DoesNotExist,'+self.server.suffix).encode('utf-8'),
            [
                (_libldap0.MOD_ADD, b'description', [b'dummy']),
            ]
        )
        self.assertIsInstance(msg_id, int)
        with self.assertRaises(_libldap0.NO_SUCH_OBJECT):
            r = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)

    def test015_modify(self):
        """
        test modify operation
        """
        ldap_conn = self._open_ldap_conn()
        # first, add an object we will delete
        dn = 'cn=AddToMe,'+self.server.suffix
        msg_id = ldap_conn.add_ext(
            dn.encode('utf-8'),
            [
                (b'objectClass', [b'person']),
                (b'cn', [b'AddToMe']),
                (b'sn', [b'Modify']),
                (b'description', [b'a description']),
            ]
        )
        self.assertIsInstance(msg_id, int)
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        self.assertEqual(result, _libldap0.RES_ADD)
        msg_id = ldap_conn.modify_ext(
            dn.encode('utf-8'),
            [(_libldap0.MOD_ADD, b'description', [b'b desc', b'c desc'])],
        )
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        self.assertEqual(result, _libldap0.RES_MODIFY)
        self.assertEqual(pmsg, [])
        self.assertEqual(msgid, msg_id)
        self.assertEqual(ctrls, [])
        # search for it back
        msg_id = ldap_conn.search_ext(
            self.server.suffix.encode('utf-8'),
            _libldap0.SCOPE_SUBTREE,
            b'(cn=AddToMe)'
        )
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        # Expect to get the objects
        self.assertEqual(result, _libldap0.RES_SEARCH_RESULT)
        self.assertEqual(len(pmsg), 1)
        self.assertEqual(msgid, msg_id)
        self.assertEqual(ctrls, [])
        self.assertEqual(pmsg[0][0], dn.encode('utf-8'))
        d = list(pmsg[0][1][b'description'])
        d.sort()
        self.assertEqual(d, [b'a description', b'b desc', b'c desc'])

    def test016_rename(self):
        ldap_conn = self._open_ldap_conn()
        dn = 'cn=RenameMe,'+self.server.suffix
        msg_id = ldap_conn.add_ext(
            dn.encode('utf-8'),
            [
                (b'objectClass', [b'organizationalRole']),
                (b'cn', [b'RenameMe']),
            ]
        )
        self.assertIsInstance(msg_id, int)
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        self.assertEqual(result, _libldap0.RES_ADD)

        # do the rename with same parent
        msg_id = ldap_conn.rename(dn.encode('utf-8'), b'cn=IAmRenamed')
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        self.assertEqual(result, _libldap0.RES_MODRDN)
        self.assertEqual(msgid, msg_id)
        self.assertEqual(pmsg, [])
        self.assertEqual(ctrls, [])

        # make sure the old one is gone
        msg_id = ldap_conn.search_ext(
            self.server.suffix.encode('utf-8'),
            _libldap0.SCOPE_SUBTREE,
            b'(cn=RenameMe)'
        )
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        self.assertEqual(result, _libldap0.RES_SEARCH_RESULT)
        self.assertEqual(len(pmsg), 0) # expect no results
        self.assertEqual(msgid, msg_id)
        self.assertEqual(ctrls, [])

        # check that the new one looks right
        dn2 = 'cn=IAmRenamed,'+self.server.suffix
        msg_id = ldap_conn.search_ext(
            self.server.suffix.encode('utf-8'),
            _libldap0.SCOPE_SUBTREE,
            b'(cn=IAmRenamed)'
        )
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        self.assertEqual(result, _libldap0.RES_SEARCH_RESULT)
        self.assertEqual(msgid, msg_id)
        self.assertEqual(ctrls, [])
        self.assertEqual(len(pmsg), 1)
        self.assertEqual(pmsg[0][0].decode('utf-8'), dn2)
        self.assertEqual(pmsg[0][1][b'cn'], [b'IAmRenamed'])

        # create the container
        containerDn = 'ou=RenameContainer,'+self.server.suffix
        msg_id = ldap_conn.add_ext(
            containerDn.encode('utf-8'),
            [
                (b'objectClass', [b'organizationalUnit']),
                (b'ou', [b'RenameContainer']),
            ]
        )
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        self.assertEqual(result, _libldap0.RES_ADD)

        # now rename from dn2 to the conater
        dn3 = 'cn=IAmRenamedAgain,' + containerDn

        # Now try renaming dn2 across container (simultaneous name change)
        msg_id = ldap_conn.rename(
            dn2.encode('utf-8'),
            b'cn=IAmRenamedAgain',
            containerDn.encode('utf-8'),
            True
        )
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        self.assertEqual(result, _libldap0.RES_MODRDN)
        self.assertEqual(msgid, msg_id)
        self.assertEqual(pmsg, [])
        self.assertEqual(ctrls, [])

        # make sure dn2 is gone
        msg_id = ldap_conn.search_ext(
            self.server.suffix.encode('utf-8'),
            _libldap0.SCOPE_SUBTREE,
            b'(cn=IAmRenamed)'
        )
        ldap_res = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        logging.debug('ldap_res = %r', ldap_res)
        result, pmsg, msgid, ctrls = ldap_res
        self.assertEqual(result, _libldap0.RES_SEARCH_RESULT)
        self.assertEqual(len(pmsg), 0) # expect no results
        self.assertEqual(msgid, msg_id)
        self.assertEqual(ctrls, [])

        msg_id = ldap_conn.search_ext(
            self.server.suffix.encode('utf-8'),
            _libldap0.SCOPE_SUBTREE,
            b'(objectClass=*)'
        )
        ldap_res = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        logging.debug('ldap_res = %r', ldap_res)
        result, pmsg, msgid, ctrls = ldap_res

        # make sure dn3 is there
        msg_id = ldap_conn.search_ext(
            self.server.suffix.encode('utf-8'),
            _libldap0.SCOPE_SUBTREE,
            b'(cn=IAmRenamedAgain)'
        )
        ldap_res = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        logging.debug('ldap_res = %r', ldap_res)
        result, pmsg, msgid, ctrls = ldap_res
        self.assertEqual(result, _libldap0.RES_SEARCH_RESULT)
        self.assertEqual(msgid, msg_id)
        self.assertEqual(ctrls, [])
        self.assertEqual(len(pmsg), 1)
        self.assertEqual(pmsg[0][0].decode('utf-8'), dn3)
        self.assertEqual(pmsg[0][1][b'cn'], [b'IAmRenamedAgain'])

    def test017_options(self):
        oldval = _libldap0.get_option(_libldap0.OPT_PROTOCOL_VERSION)
        try:
            with self.assertRaises(TypeError):
                _libldap0.set_option(_libldap0.OPT_PROTOCOL_VERSION, b'3')
            _libldap0.set_option(_libldap0.OPT_PROTOCOL_VERSION, _libldap0.VERSION2)
            v = _libldap0.get_option(_libldap0.OPT_PROTOCOL_VERSION)
            self.assertEqual(v, _libldap0.VERSION2)
            _libldap0.set_option(_libldap0.OPT_PROTOCOL_VERSION, _libldap0.VERSION3)
            v = _libldap0.get_option(_libldap0.OPT_PROTOCOL_VERSION)
            self.assertEqual(v, _libldap0.VERSION3)
        finally:
            _libldap0.set_option(_libldap0.OPT_PROTOCOL_VERSION, oldval)

        ldap_conn = self._open_ldap_conn()

        # Try changing some basic options and checking that they took effect

        ldap_conn.set_option(_libldap0.OPT_PROTOCOL_VERSION, _libldap0.VERSION2)
        v = ldap_conn.get_option(_libldap0.OPT_PROTOCOL_VERSION)
        self.assertEqual(v, _libldap0.VERSION2)

        ldap_conn.set_option(_libldap0.OPT_PROTOCOL_VERSION, _libldap0.VERSION3)
        v = ldap_conn.get_option(_libldap0.OPT_PROTOCOL_VERSION)
        self.assertEqual(v, _libldap0.VERSION3)

        # Try getting options that will yield a known error.
        with self.assertRaises(ValueError):
            _libldap0.get_option(_libldap0.OPT_MATCHED_DN)

    def test018_enotconn(self):
        ldap_conn = _libldap0._initialize(b'ldap://127.0.0.1:42')
        try:
            msg_id = ldap_conn.simple_bind(b'', b'')
            r = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)
        except _libldap0.SERVER_DOWN as ldap_err:
            err_no = ldap_err.args[0]['errno']
            if err_no != ENOTCONN:
                self.fail("expected err_no=%d, got %d" % (ENOTCONN, err_no))
        else:
            self.fail("expected SERVER_DOWN, got %r" % r)

    def test019_invalid_filter(self):
        ldap_conn = self._open_ldap_conn(bind=False)
        # search with invalid filter
        with self.assertRaises(_libldap0.FILTER_ERROR):
            msg_id = ldap_conn.search_ext(
                b'',
                _libldap0.SCOPE_BASE,
                b'(|(objectClass=*)',
            )
            self.assertIsInstance(msg_id, int)
            r = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)

    def test020_invalid_credentials(self):
        ldap_conn = self._open_ldap_conn(bind=False)
        # search with invalid filter
        with self.assertRaises(_libldap0.INVALID_CREDENTIALS):
            msg_id = ldap_conn.simple_bind(
                self.server.root_dn.encode('utf-8'),
                b'wrong_password',
            )
            r = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)

    @unittest.skip('Skipped because of unclarified timeouts')
    def test021_sasl_bind_s(self):
        ldap_conn = _libldap0._initialize(self.server.ldapi_uri.encode('ascii'))
        ldap_conn.set_option(_libldap0.OPT_PROTOCOL_VERSION, _libldap0.VERSION3)
        msg_id = ldap_conn.sasl_bind_s(b'EXTERNAL', b'', None)
        self.assertIsInstance(msg_id, int)
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)

    @unittest.skip('Skipped because of unclarified timeouts')
    def test022_sasl_interactive_bind_s(self):
        ldap_conn = _libldap0._initialize(self.server.ldapi_uri.encode('ascii'))
        ldap_conn.set_option(_libldap0.OPT_PROTOCOL_VERSION, _libldap0.VERSION3)
        msg_id = ldap_conn.sasl_interactive_bind_s(b'EXTERNAL', SaslNoninteractiveAuth(), None, 0)
        self.assertIsInstance(msg_id, int)
        result, pmsg, msgid, ctrls = ldap_conn.result(msg_id, _libldap0.MSG_ALL, self.timeout)


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.controls.sessiontrack - class for session tracking control
(see draft-wahl-ldap-session)
"""

from pyasn1.type import namedtype, univ
from pyasn1.codec.ber import encoder
from pyasn1_modules.rfc2251 import LDAPString, LDAPOID

from . import RequestControl


__all__ = [
    'SessionTrackingControl',
    'SESSION_TRACKING_CONTROL_OID',
    'SESSION_TRACKING_FORMAT_OID_RADIUS_ACCT_SESSION_ID',
    'SESSION_TRACKING_FORMAT_OID_RADIUS_ACCT_MULTI_SESSION_ID',
    'SESSION_TRACKING_FORMAT_OID_USERNAME',
]

# OID constants
SESSION_TRACKING_CONTROL_OID = '1.3.6.1.4.1.21008.108.63.1'
SESSION_TRACKING_FORMAT_OID_RADIUS_ACCT_SESSION_ID = SESSION_TRACKING_CONTROL_OID + '.1'
SESSION_TRACKING_FORMAT_OID_RADIUS_ACCT_MULTI_SESSION_ID = SESSION_TRACKING_CONTROL_OID + '.2'
SESSION_TRACKING_FORMAT_OID_USERNAME = SESSION_TRACKING_CONTROL_OID + '.3'


class SessionTrackingControl(RequestControl):
    """
    Class for Session Tracking Control

    Because criticality MUST be false for this control it cannot be set
    from the application.

    sessionSourceIp
      IP address of the request source as string
    sessionSourceName
      Name of the request source as string
    formatOID
      OID as string specifying the format
    sessionTrackingIdentifier
      String containing a specific tracking ID
    """
    controlType: str = SESSION_TRACKING_CONTROL_OID

    class SessionIdentifierControlValue(univ.Sequence):
        componentType = namedtype.NamedTypes(
            namedtype.NamedType('sessionSourceIp', LDAPString()),
            namedtype.NamedType('sessionSourceName', LDAPString()),
            namedtype.NamedType('formatOID', LDAPOID()),
            namedtype.NamedType('sessionTrackingIdentifier', LDAPString()),
        )


    def __init__(
            self,
            sessionSourceIp,
            sessionSourceName,
            formatOID,
            sessionTrackingIdentifier
        ):
        # criticality MUST be false for this control
        RequestControl.__init__(self, criticality=False)
        self.sessionSourceIp = sessionSourceIp
        self.sessionSourceName = sessionSourceName
        self.formatOID = formatOID
        self.sessionTrackingIdentifier = sessionTrackingIdentifier

    def encode(self) -> bytes:
        scv = self.SessionIdentifierControlValue()
        scv.setComponentByName(
            'sessionSourceIp',
            LDAPString(self.sessionSourceIp)
        )
        scv.setComponentByName(
            'sessionSourceName',
            LDAPString(self.sessionSourceName)
        )
        scv.setComponentByName(
            'formatOID',
            LDAPOID(self.formatOID))
        scv.setComponentByName(
            'sessionTrackingIdentifier',
            LDAPString(self.sessionTrackingIdentifier)
        )
        return encoder.encode(scv)



# -*- coding: ascii -*-
"""
ldap0.controls.ppolicy - classes for Password Policy controls
(see https://tools.ietf.org/html/draft-behera-ldap-password-policy)
"""

# Imports from pyasn1
from pyasn1.type import tag, namedtype, namedval, univ, constraint
from pyasn1.codec.ber import decoder

from . import ResponseControl
from .simple import ValueLessRequestControl


__all__ = [
    'PasswordPolicyControl'
]


class PasswordPolicyWarning(univ.Choice):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType(
            'timeBeforeExpiration',
            univ.Integer().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0)
            )
        ),
        namedtype.NamedType(
            'graceAuthNsRemaining', univ.Integer().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1)
            )
        ),
    )


class PasswordPolicyError(univ.Enumerated):
    namedValues = namedval.NamedValues(
        ('passwordExpired', 0),
        ('accountLocked', 1),
        ('changeAfterReset', 2),
        ('passwordModNotAllowed', 3),
        ('mustSupplyOldPassword', 4),
        ('insufficientPasswordQuality', 5),
        ('passwordTooShort', 6),
        ('passwordTooYoung', 7),
        ('passwordInHistory', 8)
    )
    subtypeSpec = univ.Enumerated.subtypeSpec + constraint.SingleValueConstraint(
        0, 1, 2, 3, 4, 5, 6, 7, 8
    )


class PasswordPolicyResponseValue(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.OptionalNamedType(
            'warning',
            PasswordPolicyWarning().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0)
            ),
        ),
        namedtype.OptionalNamedType(
            'error',
            PasswordPolicyError().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1)
            )
        ),
    )


class PasswordPolicyControl(ValueLessRequestControl, ResponseControl):
    controlType: str = '1.3.6.1.4.1.42.2.27.8.5.1'

    def __init__(self, criticality=False):
        ValueLessRequestControl.__init__(self, criticality=criticality)

    def decode(self, encodedControlValue: bytes):
        self.timeBeforeExpiration = None
        self.graceAuthNsRemaining = None
        self.error = None
        ppolicy_value, _ = decoder.decode(
            encodedControlValue,
            asn1Spec=PasswordPolicyResponseValue(),
        )
        warning = ppolicy_value.getComponentByName('warning')
        if warning.hasValue():
            if 'timeBeforeExpiration' in warning:
                self.timeBeforeExpiration = int(warning.getComponentByName('timeBeforeExpiration'))
            elif 'graceAuthNsRemaining' in warning:
                self.graceAuthNsRemaining = int(warning.getComponentByName('graceAuthNsRemaining'))
        error = ppolicy_value.getComponentByName('error')
        if error.hasValue():
            self.error = int(error)
        else:
            self.error = None



# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.modlist
"""

import unittest

import ldap0
import ldap0.schema.models
import ldap0.schema.util

from ldap0.modlist import add_modlist, modify_modlist

class Test001AddModlist(unittest.TestCase):

    add_modlist_tests = [
        (
            {
                'objectClass':[b'person', b'pilotPerson'],
                'cn':[b'Michael Str\303\266der', b'Michael Stroeder'],
                'sn':[b'Str\303\266der'],
                'dummy1':[],
                'dummy2':[b'2'],
                'dummy3':[b''],
            },
            [
                (b'objectClass', [b'person', b'pilotPerson']),
                (b'cn', [b'Michael Str\303\266der', b'Michael Stroeder']),
                (b'sn', [b'Str\303\266der']),
                (b'dummy2', [b'2']),
                (b'dummy3', [b'']),
            ],
            None,
        ),
        (
            {
                'objectClass':[b'person', b'pilotPerson'],
                'cn':[b'Michael Str\303\266der', b'Michael Stroeder'],
                'sn':[b'Str\303\266der'],
                'dummy1':[],
                'dummy2':[b'2'],
                'dummy3':[b''],
            },
            [
                (b'objectClass', [b'person', b'pilotPerson']),
                (b'sn', [b'Str\303\266der']),
                (b'dummy2', [b'2']),
                (b'dummy3', [b'']),
            ],
            ('cn',),
        ),
    ]

    def test_add_modlist(self):
        for entry, test_modlist, ignore_attr_types in self.add_modlist_tests:
            test_modlist.sort()
            result_modlist = add_modlist(entry, ignore_attr_types=ignore_attr_types)
            result_modlist.sort()
            self.assertEqual(
                test_modlist, result_modlist,
                'add_modlist(%s) returns\n%s\ninstead of\n%s.' % (
                    repr(entry),repr(result_modlist),repr(test_modlist)
                )
            )

class Test002ModifyModlist(unittest.TestCase):

    modify_modlist_tests = [
        (
            {
                'objectClass':[b'person', b'pilotPerson'],
                'cn':[b'Michael Str\303\266der', b'Michael Stroeder'],
                'sn':[b'Str\303\266der'],
                'enum':[b'a', b'b', b'c'],
                'c':[b'DE'],
            },
            {
                'objectClass':[b'person', b'inetOrgPerson'],
                'cn':[b'Michael Str\303\266der', b'Michael Stroeder'],
                'sn':[],
                'enum':[b'a', b'b', b'd'],
                'mail':[b'michael@stroeder.com'],
            },
            [],
            [],
            [
                (ldap0.MOD_DELETE, b'objectClass', None),
                (ldap0.MOD_ADD, b'objectClass', [b'person', b'inetOrgPerson']),
                (ldap0.MOD_DELETE, b'c', None),
                (ldap0.MOD_DELETE, b'sn', None),
                (ldap0.MOD_ADD, b'mail', [b'michael@stroeder.com']),
                (ldap0.MOD_DELETE, b'enum', None),
                (ldap0.MOD_ADD, b'enum', [b'a', b'b', b'd']),
            ]
        ),

        (
            {
                'c':[b'DE'],
            },
            {
                'c':[b'FR'],
            },
            [],
            [],
            [
                (ldap0.MOD_DELETE, b'c', None),
                (ldap0.MOD_ADD, b'c', [b'FR']),
            ]
        ),

        # Now a weird test-case for catching all possibilities
        # of removing an attribute with MOD_DELETE,attr_type, None
        (
            {
                'objectClass':[b'person'],
                'cn':[None],
                'sn':[b''],
                'c':[b'DE'],
            },
            {
                'objectClass':[],
                'cn':[],
                'sn':[None],
            },
            [],
            [],
            [
                (ldap0.MOD_DELETE, b'c', None),
                (ldap0.MOD_DELETE, b'objectClass', None),
                (ldap0.MOD_DELETE, b'sn', None),
            ]
        ),

        (
            {
                'objectClass':[b'person'],
                'cn':[b'Michael Str\303\266der', b'Michael Stroeder'],
                'sn':[b'Str\303\266der'],
                'enum':[b'a', b'b', b'C'],
            },
            {
                'objectClass':[b'Person'],
                'cn':[b'Michael Str\303\266der', b'Michael Stroeder'],
                'sn':[],
                'enum':[b'a', b'b', b'c'],
            },
            [],
            ['objectClass'],
            [
                (ldap0.MOD_DELETE, b'sn', None),
                (ldap0.MOD_DELETE, b'enum', None),
                (ldap0.MOD_ADD, b'enum', [b'a', b'b', b'c']),
            ]
        ),

        (
            {
                'objectClass':[b'person'],
                'cn':[b'Michael Str\303\266der', b'Michael Stroeder'],
                'sn':[b'Str\303\266der'],
                'enum':[b'a', b'b', b'C'],
            },
            {
                'objectClass':[b'inetOrgPerson'],
                'cn':[b'Michael Str\303\266der', b'Michael Stroeder'],
                'sn':[],
                'enum':[b'a', b'b', b'c'],
            },
            ['objectClass'],
            [],
            [
                (ldap0.MOD_DELETE, b'sn', None),
                (ldap0.MOD_DELETE, b'enum', None),
                (ldap0.MOD_ADD, b'enum', [b'a', b'b', b'c']),
            ]
        ),

    ]

    def test_modify_modlist(self):
        for old_entry, new_entry, ignore_attr_types, case_ignore_attr_types, test_modlist in self.modify_modlist_tests:
            test_modlist.sort()
            result_modlist = modify_modlist(
                old_entry, new_entry,
                ignore_attr_types=ignore_attr_types,
                case_ignore_attr_types=case_ignore_attr_types
            )
            result_modlist.sort()

            self.assertEqual(
                test_modlist, result_modlist,
                'modify_modlist(%s,%s) returns\n%s\ninstead of\n%s.' % (
                    repr(old_entry),
                    repr(new_entry),
                    repr(result_modlist),
                    repr(test_modlist),
                )
            )


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Simple tests for module ldap0.extop
"""

# from Python's standard lib
import unittest

from ldap0.extop import ExtendedRequest, ExtendedResponse
from ldap0.extop.passmod import PassmodResponse


class TestExtOpRequest(unittest.TestCase):
    """
    test classes for encoding extended operation requests
    """

    def test_basic(self):
        er = ExtendedRequest('1.2.3.4')
        self.assertEqual(er.requestName, '1.2.3.4')
        self.assertEqual(er.requestValue, None)
        self.assertEqual(repr(er), "ExtendedRequest(requestName='1.2.3.4', requestValue=None)")
        er = ExtendedRequest(requestName='1.2.3.4')
        self.assertEqual(er.requestName, '1.2.3.4')
        self.assertEqual(er.requestValue, None)
        self.assertEqual(repr(er), "ExtendedRequest(requestName='1.2.3.4', requestValue=None)")
        er = ExtendedRequest('1.2.3.4', 'foobar')
        self.assertEqual(er.requestName, '1.2.3.4')
        self.assertEqual(er.requestValue, 'foobar')
        self.assertEqual(repr(er), "ExtendedRequest(requestName='1.2.3.4', requestValue='foobar')")
        self.assertEqual(er.encode(), 'foobar')


class TestExtOpResponse(unittest.TestCase):
    """
    test classes for decoding extended operation responses
    """

    def test_basic(self):
        er = ExtendedResponse()
        self.assertEqual(er.responseValue, None)
        self.assertEqual(repr(er), "ExtendedResponse(encodedResponseValue=None)")
        er = ExtendedResponse('foobar')
        self.assertEqual(er.responseValue, 'foobar')
        self.assertEqual(repr(er), "ExtendedResponse(encodedResponseValue='foobar')")
        er.decode('foobar42')
        self.assertEqual(er.responseValue, 'foobar42')

    def test_passmod(self):
        er = PassmodResponse()
        er.decode(None)
        self.assertEqual(er.responseValue, None)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.controls for extended controls and
extended operations only usable with certain slapd overlays not
available in all OpenLDAP builds
"""

import os
import time
import unittest
import logging

# Switch off processing .ldaprc or ldap0.conf before importing _ldap
os.environ['LDAPNOINIT'] = '1'

import ldap0
from ldap0.schema.models import AttributeType
from ldap0.schema.subentry import SubSchema
from ldap0.ldapobject import LDAPObject
from ldap0.controls.ppolicy import PasswordPolicyControl
from ldap0.controls.pwdpolicy import PasswordExpiringControl, PasswordExpiredControl
from ldap0.test import SlapdTestCase
from ldap0.functions import strp_secs
from ldap0.err import (
    PasswordPolicyChangeAfterReset,
    PasswordPolicyExpirationWarning,
    PasswordPolicyExpiredError,
)


LDIF_TEMPLATE = """dn: %(suffix)s
objectClass: organization
o: %(o)s

dn: %(rootdn)s
objectClass: applicationProcess
objectClass: simpleSecurityObject
cn: %(rootcn)s
userPassword: %(rootpw)s

dn: cn=user1,%(suffix)s
objectClass: applicationProcess
objectClass: simpleSecurityObject
cn: user1
userPassword: user1_pw

"""


class SlapoPPolicyTestCase(SlapdTestCase):
    """
    base-class for tests with slapo-ppolicy
    """
    ldap_object_class = LDAPObject
    maxDiff = None
    pwd_max_age = 4
    pwd_expire_warning = 2
    pwd_graceauthn_limit = 2

    @classmethod
    def setUpClass(cls):
        super(SlapoPPolicyTestCase, cls).setUpClass()
        # insert some Foo* objects via ldapadd
        cls.server.ldapadd(
            (LDIF_TEMPLATE % {
                'suffix':cls.server.suffix,
                'rootdn':cls.server.root_dn,
                'rootcn':cls.server.root_cn,
                'rootpw':cls.server.root_pw,
                'o': cls.server.suffix.split(',')[0][2:],
            }).encode('utf-8')
        )

    def setUp(self):
        super(SlapoPPolicyTestCase, self).setUp()
        sub_schema = SubSchema(
            self._ldap_conn.read_subschemasubentry_s('cn=Subschema'),
        )
        # determine whether loading a separate schema file is needed
        # by checking presence of object class pwdPolicy
        if os.path.exists(os.path.join(self.server.SCHEMADIR, 'ppolicy.ldif')):
            slapo_ppolicy_schema = ['ppolicy.ldif']
        else:
            slapo_ppolicy_schema = None
        slapo_ppolicy_cfg = dict(
            olcOverlay=['ppolicy'],
            objectClass=['olcOverlayConfig', 'olcPPolicyConfig'],
            olcPPolicyDefault=['cn=ppolicy-default,'+self.server.suffix],
            olcPPolicyForwardUpdates=['FALSE'],
            olcPPolicyHashCleartext=['FALSE'],
            olcPPolicyUseLockout=['TRUE'],
        )
        self.enable_overlay(slapo_ppolicy_cfg, slapo_ppolicy_schema)
        # check whether attribute type olcPPolicySendNetscapeControls is present in subschema
        sub_schema = SubSchema(
            self._ldap_conn.read_subschemasubentry_s('cn=Subschema'),
        )
        self.send_netscape_controls = ('1.3.6.1.4.1.4203.1.12.2.3.3.12.6' in sub_schema.sed[AttributeType])
        if self.send_netscape_controls:
            # enable ppolicy_send_send_netscape_controls available since OpenLDAP ITS#9279
            slapo_ppolicy_cfg.update(olcPPolicySendNetscapeControls=['TRUE'])
            self._ldap_conn.modify_s(
                'olcOverlay={0}ppolicy,olcDatabase={1}mdb,cn=config',
                [(ldap0.MOD_REPLACE, b'olcPPolicySendNetscapeControls', [b'TRUE'])]
            )
            logging.debug('Set olcPPolicySendNetscapeControls: TRUE')
        try:
            self._add_ppolicy_entry(
                'default',
                pwdMaxAge=self.pwd_max_age,
                pwdExpireWarning=self.pwd_expire_warning,
                pwdGraceAuthNLimit=self.pwd_graceauthn_limit,
                pwdAllowUserChange='TRUE',
                pwdMustChange='TRUE',
            )
        except ldap0.ALREADY_EXISTS:
            pass
        # end of setUp()

    @property
    def user_dn(self):
        return 'cn=user1,' + self.server.suffix

    def _add_ppolicy_entry(self, name, **ppolicy_attrs):
        """
        add a new password policy entry
        """
        logging.debug('ppolicy_attrs = %r', ppolicy_attrs)
        ppolicy_cn = 'ppolicy-%s' % (name)
        ppolicy_dn = 'cn=%s,%s' % (ppolicy_cn, self.server.suffix)
        ppolicy_entry = {
            'objectClass': [b'applicationProcess', b'pwdPolicy'],
            'cn': [ppolicy_cn.encode('utf-8')],
            'pwdAttribute': [b'userPassword'],
        }
        for at, av in ppolicy_attrs.items():
            if av is not None:
                ppolicy_entry[at] = [str(av).encode('utf-8')]
        self._ldap_conn.add_s(
            ppolicy_dn,
            ppolicy_entry,
        )
        read_entry = self._ldap_conn.read_s(
            ppolicy_dn,
            attrlist=list(ppolicy_entry.keys()),
        ).entry_as
        self.assertEqual(read_entry, ppolicy_entry)

    def _set_user_password(self, user_password, pwd_reset=None):
        modlist = [(ldap0.MOD_REPLACE, b'userPassword', [user_password.encode('utf-8')])]
        if pwd_reset is not None:
            modlist.append(
                (
                    ldap0.MOD_REPLACE,
                    b'pwdReset',
                    [str(pwd_reset).upper().encode('ascii')]
                )
            )
        self._ldap_conn.modify_s(self.user_dn, modlist)
        return self._ldap_conn.read_s(
            self.user_dn,
            attrlist=['pwdChangedTime', 'pwdReset'],
        )


class TestPPolicy(SlapoPPolicyTestCase):
    """
    tests for ldap0.controls.ppolicy
    """

    def test001_ppolicy_reset(self):
        user_password = 'supersecret1'
        user = self._set_user_password(user_password, pwd_reset=True)
        self.assertEqual(user.entry_s['pwdReset'], ['TRUE'])
        # check a normal simple bind first
        with self.ldap_object_class(self.server.ldap_uri, trace_level=self.trace_level) as l:
            with self.assertRaises(PasswordPolicyChangeAfterReset) as ctx:
                l.simple_bind_s(
                    self.user_dn,
                    user_password.encode('utf-8'),
                    req_ctrls=[PasswordPolicyControl()],
                )
            self.assertEqual(str(ctx.exception), 'Password change is needed after reset!')
            # let the user change his own password
            l.passwd_s(self.user_dn, None, 'supersecret2')
        # end of test001_pwd_reset()

    def test002_ppolicy_expiry_warning(self):
        user_password = b'supersecret2'
        user = self._ldap_conn.read_s(
            self.user_dn,
            attrlist=['pwdChangedTime', 'pwdReset'],
        )
        pwd_changed_time = strp_secs(user.entry_s['pwdChangedTime'][0])
        self.assertTrue('pwdReset' not in user.entry_s)
        # wait until grace period and check for PasswordPolicyExpirationWarning
        sleep_time = pwd_changed_time + self.pwd_max_age - self.pwd_expire_warning + 1.0 - time.time()
        logging.debug('Wait for %s seconds', sleep_time)
        time.sleep(sleep_time)
        with self.ldap_object_class(self.server.ldap_uri, trace_level=self.trace_level) as l:
            with self.assertRaises(PasswordPolicyExpirationWarning) as ctx:
                l.simple_bind_s(
                    self.user_dn,
                    user_password,
                    req_ctrls=[PasswordPolicyControl()],
                )
            expected_time_left = int(self.pwd_max_age-self.pwd_expire_warning-1)
            self.assertEqual(ctx.exception.timeBeforeExpiration, expected_time_left)
            self.assertEqual(
                str(ctx.exception),
                'Password will expire in %d seconds!' % (expected_time_left,),
            )
        # end of test002_pwd_expiry_warning()

    def test003_ppolicy_grace_logins(self):
        user_password = 'supersecret2'
        user_entry = self._ldap_conn.read_s(
            self.user_dn,
            attrlist=['pwdChangedTime', 'pwdReset'],
        ).entry_s
        pwd_changed_time = strp_secs(user_entry['pwdChangedTime'][0])
        self.assertTrue('pwdReset' not in user_entry)
        # wait until password expired and check for PasswordPolicyExpiredError
        sleep_time = pwd_changed_time + self.pwd_max_age + 1.0 - time.time()
        logging.debug('Wait for %s seconds', sleep_time)
        time.sleep(sleep_time)
        # change in slapo-ppolicy 2.5.x (see ITS#7596)
        ol25_offset = int(not self.server.slapd_version.startswith('2.4.'))
        for pwd_graceauthn_limit in range(self.pwd_graceauthn_limit-ol25_offset, -ol25_offset, -1):
            with self.ldap_object_class(self.server.ldap_uri, trace_level=self.trace_level) as l:
                with self.assertRaises(PasswordPolicyExpiredError) as ctx:
                    l.simple_bind_s(
                        self.user_dn,
                        user_password.encode('utf-8'),
                        req_ctrls=[PasswordPolicyControl()],
                    )
                self.assertEqual(
                    str(ctx.exception),
                    'Password expired! {0:d} grace logins left.'.format(pwd_graceauthn_limit)
                )
                self.assertEqual(ctx.exception.graceAuthNsRemaining, pwd_graceauthn_limit)
                del ctx
                # OpenLDAP's slapd derives graceAuthNsRemaining from number
                # of pwdGraceUseTime values, which are only stored with
                # a granularity of a second
                # => sleep for a bit more than second to be on the safe side
                time.sleep(1.1)
        with self.ldap_object_class(self.server.ldap_uri, trace_level=self.trace_level) as l:
            with self.assertRaises(ldap0.INVALID_CREDENTIALS) as ctx:
                l.simple_bind_s(
                    self.user_dn,
                    user_password.encode('utf-8'),
                    req_ctrls=[PasswordPolicyControl()],
                )
            self.assertEqual(len(ctx.exception.ctrls), 1)
            self.assertIsInstance(ctx.exception.ctrls[0], PasswordPolicyControl)
            self.assertEqual(ctx.exception.ctrls[0].error, 0)
        # end of test003_pwd_grace_logins()


class TestPwdPolicy(SlapoPPolicyTestCase):
    """
    tests for ldap0.controls.pwdpolicy
    """
    pwd_max_age = 5
    pwd_expire_warning = 3
    pwd_graceauthn_limit = None

    def setUp(self):
        super(TestPwdPolicy, self).setUp()
        # FIX ME! Running these tests on OpenLDAP 2.5+ needs clarification.
        if not self.server.slapd_version.startswith('2.4.'):
            self.skipTest(
                '{} known to fail with slapd {}'.format(
                    self.__class__.__name__,
                    self.server.slapd_version,
                )
            )

    def test001_pwdpolicy_expiration(self):
        if not self.send_netscape_controls:
            self.skipTest('slapd has no support for ITS#9279')
        user_password = 'supersecret2'
        user = self._set_user_password(user_password)
        user = self._ldap_conn.read_s(
            self.user_dn,
            attrlist=['pwdChangedTime'],
        )
        pwd_changed_time = strp_secs(user.entry_s['pwdChangedTime'][0])
        # wait until grace period and check for PasswordPolicyExpirationWarning
        sleep_time = pwd_changed_time + self.pwd_max_age - self.pwd_expire_warning + 1.0 - time.time()
        logging.debug('Wait for %s seconds', sleep_time)
        time.sleep(sleep_time)
        with self.ldap_object_class(self.server.ldap_uri, trace_level=self.trace_level) as l:
            bind_res = l.simple_bind_s(self.user_dn, user_password.encode('utf-8'))
            logging.debug('bind_res = %r', bind_res)
        self.assertEqual(len(bind_res.ctrls), 1)
        self.assertIsInstance(bind_res.ctrls[0], PasswordExpiringControl)
        self.assertTrue(
            bind_res.ctrls[0].gracePeriod <= self.pwd_expire_warning
        )
        # end of test004_pwdpolicy_expiration()

    def test002_pwdpolicy_expired(self):
        if not self.send_netscape_controls:
            self.skipTest('slapd has no support for ITS#9279')
        user_password = 'supersecret2'
        user = self._ldap_conn.read_s(
            self.user_dn,
            attrlist=['pwdChangedTime'],
        )
        pwd_changed_time = strp_secs(user.entry_s['pwdChangedTime'][0])
        # wait until grace period and check for PasswordPolicyExpirationWarning
        sleep_time = pwd_changed_time + self.pwd_max_age + 1.0 - time.time()
        logging.debug('Wait for %s seconds', sleep_time)
        time.sleep(sleep_time)
        with self.ldap_object_class(self.server.ldap_uri, trace_level=self.trace_level) as l:
            with self.assertRaises(ldap0.INVALID_CREDENTIALS) as ctx:
                l.simple_bind_s(self.user_dn, user_password.encode('utf-8'))
            self.assertEqual(len(ctx.exception.ctrls), 1)
            self.assertIsInstance(ctx.exception.ctrls[0], PasswordExpiredControl)
            self.assertTrue(ctx.exception.ctrls[0].passwordExpired)
        # end of test005_pwdpolicy_expired()


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.openldap -- OpenLDAP-specific stuff
"""

from typing import Optional, Set

from .ldapurl import LDAPUrl
from .ldapurl import SEARCH_SCOPE as LDAPURL_SEARCH_SCOPE
from .functions import attr_set

__all__ = [
    'SyncReplDesc',
    'ldapsearch_cmd',
]

LDAPSEARCH_TMPL = (
    'ldapsearch '
    '-LL '
    '-H "{uri}" '
    '{tls} '
    '-b "{dn}" '
    '-s {scope} '
    '{authc} '
    '{authz} '
    '"{filterstr}" '
    '{attrs}'
)


def ldapsearch_cmd(ldap_url: LDAPUrl) -> str:
    """
    Returns string with OpenLDAP compatible ldapsearch command based on LDAPUrl instance.
    """
    if ldap_url.attrs is None:
        attrs_str = ''
    else:
        attrs_str = ' '.join(ldap_url.attrs)
    scope_str = {
        0: 'base',
        1: 'one',
        2: 'sub',
        3: 'children',
    }[ldap_url.scope]
    if ldap_url.sasl_mech:
        authc_str = '-Y "{saslmech}"'.format(saslmech=ldap_url.sasl_mech)
    elif ldap_url.who:
        authc_str = '-x -D "{who}" -W'.format(who=ldap_url.who or '')
    else:
        authc_str = '-x -D "" -w ""'
    if ldap_url.start_tls:
        tls_str = '-ZZ'
    else:
        tls_str = ''
    if ldap_url.sasl_authzid:
        authz_str = '-X "{authz}"'.format(authz=ldap_url.sasl_authzid)
    else:
        authz_str = ''
    if ldap_url.extensions:
        # FIX ME! Set extended controls
        pass
    return LDAPSEARCH_TMPL.format(
        uri=ldap_url.connect_uri(),
        dn=ldap_url.dn,
        scope=scope_str,
        attrs=attrs_str,
        filterstr=ldap_url.filterstr or '(objectClass=*)',
        authc=authc_str,
        authz=authz_str,
        tls=tls_str,
    ).strip()
    # end of ldapsearch_cmd()


class SyncReplDesc:
    """
    Parser class for OpenLDAP syncrepl directives
    """
    known_keywords = (
        'attrs',
        'attrsonly',
        'authcid',
        'authzid',
        'binddn',
        'bindmethod',
        'credentials',
        'exattrs',
        'filter',
        'interval',
        'keepalive',
        'logbase',
        'logfilter',
        'network-timeout',
        'provider',
        'realm',
        'retry',
        'rid',
        'saslmech',
        'schemachecking',
        'scope',
        'searchbase',
        'secprops',
        'sizelimit',
        'starttls',
        'suffixmassage',
        'syncdata',
        'timelimit',
        'timeout',
        'tls_cacert',
        'tls_cacertdir',
        'tls_cert',
        'tls_cipher_suite',
        'tls_crlcheck',
        'tls_key',
        'tls_reqcert',
        'tls_protocol_min',
        'tls_reqsan',
        'type',
    )
    enum_args = {
        'bindmethod': {'simple', 'sasl'},
        'scope': {'sub', 'one', 'base', 'subord'},
        'starttls': {'no', 'demand', 'critical'},
        'syncdata': {'default', 'accesslog', 'changelog'},
        'tls_crlcheck': {'none', 'peer', 'all'},
        'tls_reqcert': {'never', 'allow', 'try', 'demand'},
        'tls_reqsan': {'never', 'allow', 'try', 'demand'},
        'type': {'refreshonly', 'refreshandpersist'},
    }
    sanitize_func = {
        'attrs': attr_set,
        'exattrs': attr_set,
        'timeout': int,
        'network-timeout': int,
        'sizelimit': int,
        'timelimit': int,
        'bindmethod': str.lower,
        'saslmech': str.upper,
    }
    attrs: Optional[Set[str]]
    attrsonly: Optional[str]
    authcid: Optional[str]
    authzid: Optional[str]
    binddn: Optional[str]
    bindmethod: Optional[str]
    credentials: Optional[str]
    exattrs: Optional[Set[str]]
    filter: Optional[str]
    interval: Optional[str]
    keepalive: Optional[str]
    logbase: Optional[str]
    logfilter: Optional[str]
    network_timeout: Optional[int]
    provider: Optional[str]
    realm: Optional[str]
    retry: Optional[str]
    rid: Optional[str]
    saslmech: Optional[str]
    schemachecking: Optional[str]
    scope: Optional[int]
    searchbase: Optional[str]
    secprops: Optional[str]
    sizelimit: Optional[int]
    starttls: Optional[str]
    suffixmassage: Optional[str]
    syncdata: Optional[str]
    timelimit: Optional[int]
    timeout: Optional[int]
    tls_cacert: Optional[str]
    tls_cacertdir: Optional[str]
    tls_cert: Optional[str]
    tls_cipher_suite: Optional[str]
    tls_crlcheck: Optional[str]
    tls_key: Optional[str]
    tls_reqcert: Optional[str]
    tls_protocol_min: Optional[str]
    type: Optional[str]

    def __init__(self, syncrepl_statement: str):
        """
        syncrepl_statement
           syncrepl statement without any line breaks
        """
        # strip all white spaces from syncrepl statement parameters
        syncrepl_statement = syncrepl_statement.strip()
        # Set class attributes for all known keywords
        for keyword in self.known_keywords:
            setattr(self, keyword.replace('-', '_'), None)
        parts = []
        for keyword in self.known_keywords:
            k_pos = syncrepl_statement.find(keyword+'=')
            if k_pos == 0 or (k_pos > 0 and syncrepl_statement[k_pos-1] == ' '):
                parts.append(k_pos)
        parts.sort()
        parts_len = len(parts)
        for ind in range(parts_len):
            if ind == parts_len-1:
                next_pos = len(syncrepl_statement)
            else:
                next_pos = parts[ind+1]
            key, val = syncrepl_statement[parts[ind]:next_pos].split('=', 1)
            key = key.strip()
            val = val.strip()
            if val[0] == '"' and val[-1] == '"':
                val = val[1:-1]
            if key in self.sanitize_func:
                val = self.sanitize_func[key](val)
            if key in self.enum_args and val.lower() not in self.enum_args[key]:
                raise ValueError('%r not in %r, was %r' % (
                    key, self.enum_args[key], val
                ))
            setattr(self, key.replace('-', '_'), val)
        # transform scope
        if self.scope is not None:
            self.scope = LDAPURL_SEARCH_SCOPE[self.scope]

    def __repr__(self) -> str:
        return '%s(rid=%s)' % (self.__class__.__name__, self.rid)

    def ldap_url(self) -> LDAPUrl:
        """
        Return ldapurl.LDAPUrl object representing some syncrepl parameters
        as close as possible.
        """
        ldap_url = LDAPUrl(self.provider)
        ldap_url.dn = self.searchbase
        ldap_url.scope = self.scope
        ldap_url.filterstr = self.filter
        ldap_url.who = self.authcid or self.binddn
        ldap_url.cred = self.credentials
        ldap_url.attrs = list(self.attrs or ['*', '+'])
        return ldap_url



# -*- coding: ascii -*-
"""
ldap0.extop.whoami - Classes for "Who am I?" extended operation
(see RFC 4532)
"""

from . import ExtendedRequest, ExtendedResponse


class WhoAmIRequest(ExtendedRequest):
    """
    Who Am I? extended request
    """

    requestName: str = '1.3.6.1.4.1.4203.1.11.3'


class WhoAmIResponse(ExtendedResponse):
    """
    Who Am I? extended response
    """

    responseName = None

    @classmethod
    def check_resp_name(cls, name):
        """
        work around MS AD returning obscure OID as responseName
        """
        return name in (cls.responseName, WhoAmIRequest.requestName)

    def decode(self, value: bytes):
        self.responseValue = value.decode(self.encoding, errors='strict')



# -*- coding: ascii -*-
"""
ldap0 - base module
"""

import _libldap0
from _libldap0 import *

from .__about__ import __version__, __author__, __license__

# Tracing is only supported in debugging mode
_trace_level = 0

import _libldap0
if _libldap0.__version__.decode('ascii') != __version__:
    raise ImportError(
        'ldap0 %s and _libldap0 %s version mismatch!' % (
            __version__,
            _libldap0.__version__.decode('ascii'),
        )
    )

# call into libldap to initialize it right now
LIBLDAP_API_INFO = get_option(_libldap0.OPT_API_INFO)
# set reasonable global defaults for various options
set_option(_libldap0.OPT_PROTOCOL_VERSION, _libldap0.VERSION3)
set_option(_libldap0.OPT_RESTART, False)
set_option(_libldap0.OPT_DEREF, False)
set_option(_libldap0.OPT_REFERRALS, False)
set_option(_libldap0.OPT_X_TLS_REQUIRE_CERT, _libldap0.OPT_X_TLS_HARD)
# name of crypto library libldap is linked with
LIBLDAP_TLS_PACKAGE = _libldap0.get_option(_libldap0.OPT_X_TLS_PACKAGE).decode('ascii')

OPT_NAMES = {
    val: key
    for key, val in vars(_libldap0).items()
    if key.startswith('OPT_')
}



# -*- coding: ascii -*-
"""
ldap0.schema.util -  misc. subschema utility stuff
"""

from urllib.request import urlopen

from _libldap0 import MOD_DELETE, MOD_ADD

from ..cidict import CIDict
from ..ldapurl import LDAPUrl, is_ldapurl
from ..ldapobject import LDAPObject
from ..ldif import LDIFParser

from ..schema import SubSchema, SCHEMA_ATTRS, SCHEMA_CLASS_MAPPING
from .models import AttributeType, SchemaElementOIDSet
from . import subentry, models


def urlfetch(uri, trace_level=0, check_uniqueness=True, rootdse=True):
    """
    Fetches a parsed schema entry by uri.

    If :uri: is a LDAP URL the LDAP server is queried directly.
    Otherwise :uri: is assumed to point to a LDIF file which
    is loaded with urllib.

    :trace_level: is passed to class LDAPObject for debugging.

    :check_uniqueness: is passed to SubSchema().

    :rootdse: is passed to LDAPObject.search_subschemasubentry_s()
    """
    uri = uri.strip()
    if is_ldapurl(uri):
        ldap_url = LDAPUrl(uri)
        ldap_conn = LDAPObject(ldap_url, trace_level)
        ldap_conn.simple_bind_s(
            ldap_url.who or '',
            (ldap_url.cred or '').encode(ldap_conn.encoding),
        )
        subschemasubentry_dn = ldap_conn.search_subschemasubentry_s(
            ldap_url.dn,
            rootdse=rootdse,
        )
        if subschemasubentry_dn is None:
            s_temp = None
        else:
            if ldap_url.attrs is None:
                schema_attrs = SCHEMA_ATTRS
            else:
                schema_attrs = ldap_url.attrs
            s_temp = ldap_conn.read_subschemasubentry_s(subschemasubentry_dn, attrs=schema_attrs)
        ldap_conn.unbind_s()
        del ldap_conn
    else:
        with urlopen(uri) as ldif_file:
            ldif_parser = LDIFParser(ldif_file, max_entries=1)
            ldif_res = ldif_parser.list_entry_records()
        subschemasubentry_dn, s_temp = ldif_res[0]
        s_temp = {
            at.decode('ascii'): [av.decode('utf-8') for av in avs]
            for at, avs in (s_temp or {}).items()
        }
    # Work-around for mixed-cased schema attribute names
    subschemasubentry_entry = CIDict()
    subschemasubentry_entry.update({
        at: avs
        for at, avs in s_temp.items()
        if at in SCHEMA_CLASS_MAPPING
    })
    del s_temp
    # Finally parse the schema
    if subschemasubentry_dn is not None:
        parsed_sub_schema = SubSchema(
            subschemasubentry_entry,
            check_uniqueness=check_uniqueness,
        )
    else:
        parsed_sub_schema = None
    return subschemasubentry_dn, parsed_sub_schema


def modify_modlist(
        sub_schema,
        old_entry,
        new_entry,
        ignore_attr_types=None,
        ignore_oldexistent=0,
        order_max=3,
    ):
    """
    Build differential modify list for calling LDAPObject.modify()/modify_s()
    with schema knowledge

    sub_schema
        Instance of ldaputil.schema.SubSchema
    old_entry
        Dictionary holding the old entry
    new_entry
        Dictionary holding what the new entry should be
    ignore_attr_types
        Sequence of attribute type names to be ignored completely
    ignore_oldexistent
        If non-zero attribute type names which are in old_entry
        but are not found in new_entry at all are not deleted.
        This is handy for situations where your application
        sets attribute value to '' for deleting an attribute.
        In most cases leave zero.
    order_max
        Max. number of attribute values to be reordered to match
        order in new_entry
    """
    # Type checking
    assert isinstance(sub_schema, subentry.SubSchema), TypeError(
        'Expected sub_schema to be instance of SubSchema, got %r' % (sub_schema,)
    )
    assert isinstance(old_entry, models.Entry), TypeError(
        'Expected old_entry to be instance of models.Entry, got %r' % (old_entry,)
    )
    assert isinstance(new_entry, models.Entry), TypeError(
        'Expected new_entry to be instance of models.Entry, got %r' % (new_entry,)
    )

    if isinstance(ignore_attr_types, SchemaElementOIDSet):
        ignore_attr_type_set = ignore_attr_types
    else:
        ignore_attr_type_set = SchemaElementOIDSet(
            sub_schema,
            AttributeType,
            ignore_attr_types,
        )

    # Performance optimization
    at_cls = models.AttributeType
    mr_cls = models.MatchingRule

    # Start building the modlist result
    modlist = []

    # Sanitize new_entry
    for attr_type in list(new_entry):
        # Filter away list items which are empty strings or None
        new_entry[attr_type] = [av for av in new_entry[attr_type] if av]
        # Check for attributes with empty value lists
        if not new_entry[attr_type]:
            # Remove the empty attribute
            del new_entry[attr_type]

    for attr_type in new_entry.keys():

        attr_type_b = attr_type.encode('ascii')

        if attr_type in ignore_attr_type_set:
            # This attribute type is ignored
            continue

        # Check whether there's an equality matching rule defined
        # for the attribute type and the matching rule is announced in subschema
        try:
            at_eq_mr = sub_schema.get_inheritedattr(
                at_cls,
                attr_type,
                'equality',
            )
        except KeyError:
            mr_obj = None
        else:
            if at_eq_mr:
                mr_obj = sub_schema.get_obj(mr_cls, at_eq_mr)
            else:
                mr_obj = None

        # Filter away list items which are empty strings or None
        new_value = new_entry[attr_type]
        old_value = [av for av in old_entry.get(attr_type, []) if av]

        # We have to check if attribute value lists differs
        old_value_set = set(old_value)
        new_value_set = set(new_value)

        if mr_obj:
            # for attributes with equality matching rule try to
            # generate a fine-grained diff
            if len(old_value) > order_max or len(new_value) > order_max:
                # for "many" values be less invasive but not order-preserving
                del_values = [v for v in old_value if not v in new_value_set]
                add_values = [v for v in new_value if not v in old_value_set]
                if old_value and del_values:
                    modlist.append((MOD_DELETE, attr_type_b, del_values))
                if new_value and add_values:
                    modlist.append((MOD_ADD, attr_type_b, add_values))
            else:
                # for few values be order-preserving
                if new_value and old_value != new_value:
                    if old_value:
                        modlist.append((MOD_DELETE, attr_type_b, old_value))
                    modlist.append((MOD_ADD, attr_type_b, new_value))
        elif new_value and old_value != new_value:
            if old_value:
                modlist.append((MOD_DELETE, attr_type_b, None))
            modlist.append((MOD_ADD, attr_type_b, new_value))

    # Remove all attributes of old_entry which are not present
    # in new_entry at all
    if not ignore_oldexistent:
        for attr_type, old_values in old_entry.items():

            attr_type_b = attr_type.encode('ascii')

            if (
                    old_values and
                    old_values != [b''] and
                    not attr_type in new_entry and
                    not sub_schema.get_oid(at_cls, attr_type) in ignore_attr_type_set
                ):
                try:
                    at_eq_mr = sub_schema.get_inheritedattr(
                        at_cls,
                        attr_type,
                        'equality',
                    )
                except KeyError:
                    at_eq_mr = None
                if at_eq_mr:
                    mr_obj = sub_schema.get_obj(mr_cls, at_eq_mr)
                    if mr_obj:
                        modlist.append((MOD_DELETE, attr_type_b, old_values))
                    else:
                        modlist.append((MOD_DELETE, attr_type_b, None))
                else:
                    modlist.append((MOD_DELETE, attr_type_b, None))

    return modlist
    # end of modify_modlist()
